"""This module defines the IndexSum class."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from ufl.constantvalue import Zero
from ufl.core.expr import Expr, ufl_err_str
from ufl.core.multiindex import MultiIndex
from ufl.core.operator import Operator
from ufl.core.ufl_type import ufl_type
from ufl.precedence import parstr

# --- Sum over an index ---


@ufl_type(num_ops=2)
class IndexSum(Operator):
    """Index sum."""

    __slots__ = ("_dimension", "ufl_free_indices", "ufl_index_dimensions")

    def __new__(cls, summand, index):
        """Create a new IndexSum."""
        # Error checks
        if not isinstance(summand, Expr):
            raise ValueError(f"Expecting Expr instance, got {ufl_err_str(summand)}")
        if not isinstance(index, MultiIndex):
            raise ValueError(f"Expecting MultiIndex instance, got {ufl_err_str(index)}")
        if len(index) != 1:
            raise ValueError(f"Expecting a single Index but got {len(index)}.")

        # Simplification to zero
        if isinstance(summand, Zero):
            sh = summand.ufl_shape
            (j,) = index
            fi = summand.ufl_free_indices
            fid = summand.ufl_index_dimensions
            pos = fi.index(j.count())
            fi = fi[:pos] + fi[pos + 1 :]
            fid = fid[:pos] + fid[pos + 1 :]
            return Zero(sh, fi, fid)

        return Operator.__new__(cls)

    def __init__(self, summand, index):
        """Initialise."""
        (j,) = index
        fi = summand.ufl_free_indices
        fid = summand.ufl_index_dimensions
        pos = fi.index(j.count())
        self._dimension = fid[pos]
        self.ufl_free_indices = fi[:pos] + fi[pos + 1 :]
        self.ufl_index_dimensions = fid[:pos] + fid[pos + 1 :]
        Operator.__init__(self, (summand, index))

    def index(self):
        """Get index."""
        return self.ufl_operands[1][0]

    def dimension(self):
        """Get dimension."""
        return self._dimension

    @property
    def ufl_shape(self):
        """Get UFL shape."""
        return self.ufl_operands[0].ufl_shape

    def evaluate(self, x, mapping, component, index_values):
        """Evaluate."""
        (i,) = self.ufl_operands[1]
        tmp = 0
        for k in range(self._dimension):
            index_values.push(i, k)
            tmp += self.ufl_operands[0].evaluate(x, mapping, component, index_values)
            index_values.pop()
        return tmp

    def __str__(self):
        """Format as a string."""
        return "sum_{%s} %s " % (str(self.ufl_operands[1]), parstr(self.ufl_operands[0], self))



"""Various sequence manipulation utilities."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from functools import reduce

import numpy as np


def product(sequence):
    """Return the product of all elements in a sequence."""
    p = 1
    for f in sequence:
        p *= f
    return p


def max_degree(degrees):
    """Maximum degree for mixture of scalar and tuple degrees."""
    # np.maximum broadcasts scalar degrees to tuple degrees if
    # necessary.  reduce applies np.maximum pairwise.
    degree = reduce(np.maximum, map(np.asarray, degrees))
    if degree.ndim:
        degree = tuple(map(int, degree))  # tuple degree
    else:
        degree = int(degree)  # scalar degree
    return degree



"""Test FFC forms.

Unit tests including all demo forms from FFC 0.5.0. The forms are
modified (with comments) to work with the UFL notation which differs
from the FFC notation in some places.
"""

__author__ = "Anders Logg (logg@simula.no) et al."
__date__ = "2008-04-09 -- 2008-09-26"
__copyright__ = "Copyright (C) 2008 Anders Logg et al."
__license__ = "GNU GPL version 3 or any later version"

# Examples copied from the FFC demo directory, examples contributed
# by Johan Jansson, Kristian Oelgaard, Marie Rognes, and Garth Wells.

from ufl import (
    Coefficient,
    Constant,
    Dx,
    FacetNormal,
    FunctionSpace,
    Mesh,
    TestFunction,
    TestFunctions,
    TrialFunction,
    TrialFunctions,
    VectorConstant,
    avg,
    curl,
    div,
    dot,
    dS,
    ds,
    dx,
    grad,
    i,
    inner,
    j,
    jump,
    lhs,
    rhs,
    sqrt,
    tetrahedron,
    triangle,
)
from ufl.finiteelement import FiniteElement, MixedElement
from ufl.pullback import contravariant_piola, covariant_piola, identity_pullback
from ufl.sobolevspace import H1, L2, HCurl, HDiv


def testConstant():
    element = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)

    v = TestFunction(space)
    u = TrialFunction(space)

    c = Constant(domain)
    d = VectorConstant(domain)

    _ = c * dot(grad(v), grad(u)) * dx

    # FFC notation: L = dot(d, grad(v))*dx
    _ = inner(d, grad(v)) * dx


def testElasticity():
    element = FiniteElement("Lagrange", tetrahedron, 1, (3,), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", tetrahedron, 1, (3,), identity_pullback, H1))
    space = FunctionSpace(domain, element)

    v = TestFunction(space)
    u = TrialFunction(space)

    def eps(v):
        # FFC notation: return grad(v) + transp(grad(v))
        return grad(v) + (grad(v)).T

    # FFC notation: a = 0.25*dot(eps(v), eps(u))*dx
    _ = 0.25 * inner(eps(v), eps(u)) * dx


def testEnergyNorm():
    element = FiniteElement("Lagrange", tetrahedron, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", tetrahedron, 1, (3,), identity_pullback, H1))
    space = FunctionSpace(domain, element)

    v = Coefficient(space)
    _ = (v * v + dot(grad(v), grad(v))) * dx


def testEquation():
    element = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)

    k = 0.1

    v = TestFunction(space)
    u = TrialFunction(space)
    u0 = Coefficient(space)

    F = v * (u - u0) * dx + k * dot(grad(v), grad(0.5 * (u0 + u))) * dx

    _ = lhs(F)
    _ = rhs(F)


def testFunctionOperators():
    element = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)

    v = TestFunction(space)
    u = TrialFunction(space)
    f = Coefficient(space)
    g = Coefficient(space)

    # FFC notation: a = sqrt(1/modulus(1/f))*sqrt(g)*dot(grad(v), grad(u))*dx
    # + v*u*sqrt(f*g)*g*dx
    _ = sqrt(1 / abs(1 / f)) * sqrt(g) * dot(grad(v), grad(u)) * dx + v * u * sqrt(f * g) * g * dx


def testHeat():
    element = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)

    v = TestFunction(space)
    u1 = TrialFunction(space)
    u0 = Coefficient(space)
    c = Coefficient(space)
    f = Coefficient(space)
    k = Constant(domain)

    _ = v * u1 * dx + k * c * dot(grad(v), grad(u1)) * dx
    _ = v * u0 * dx + k * v * f * dx


def testMass():
    element = FiniteElement("Lagrange", tetrahedron, 3, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", tetrahedron, 1, (3,), identity_pullback, H1))
    space = FunctionSpace(domain, element)
    v = TestFunction(space)
    u = TrialFunction(space)
    _ = v * u * dx


def testMixedMixedElement():
    P3 = FiniteElement("Lagrange", triangle, 3, (), identity_pullback, H1)
    MixedElement([[P3, P3], [P3, P3]])


def testMixedPoisson():
    q = 1
    BDM = FiniteElement("Brezzi-Douglas-Marini", triangle, q, (2,), contravariant_piola, HDiv)
    DG = FiniteElement("Discontinuous Lagrange", triangle, q - 1, (), identity_pullback, L2)

    mixed_element = MixedElement([BDM, DG])
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, mixed_element)

    (tau, w) = TestFunctions(space)
    (sigma, u) = TrialFunctions(space)
    f = Coefficient(FunctionSpace(domain, DG))
    _ = (dot(tau, sigma) - div(tau) * u + w * div(sigma)) * dx
    _ = w * f * dx


def testNavierStokes():
    element = FiniteElement("Lagrange", tetrahedron, 1, (3,), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", tetrahedron, 1, (3,), identity_pullback, H1))
    space = FunctionSpace(domain, element)

    v = TestFunction(space)
    u = TrialFunction(space)
    w = Coefficient(space)

    # FFC notation: a = v[i]*w[j]*D(u[i], j)*dx
    _ = v[i] * w[j] * Dx(u[i], j) * dx


def testNeumannProblem():
    element = FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)

    v = TestFunction(space)
    u = TrialFunction(space)
    f = Coefficient(space)
    g = Coefficient(space)

    # FFC notation: a = dot(grad(v), grad(u))*dx
    _ = inner(grad(v), grad(u)) * dx

    # FFC notation: L = dot(v, f)*dx + dot(v, g)*ds
    _ = inner(v, f) * dx + inner(v, g) * ds


def testOptimization():
    element = FiniteElement("Lagrange", triangle, 3, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)
    v = TestFunction(space)
    u = TrialFunction(space)
    f = Coefficient(space)
    _ = dot(grad(v), grad(u)) * dx
    _ = v * f * dx


def testP5tet():
    FiniteElement("Lagrange", tetrahedron, 5, (), identity_pullback, H1)


def testP5tri():
    FiniteElement("Lagrange", triangle, 5, (), identity_pullback, H1)


def testPoissonDG():
    element = FiniteElement("Discontinuous Lagrange", triangle, 1, (), identity_pullback, L2)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)

    v = TestFunction(space)
    u = TrialFunction(space)
    f = Coefficient(space)
    n = FacetNormal(domain)

    # FFC notation: h = MeshSize(domain), not supported by UFL
    h = Constant(domain)

    gN = Coefficient(space)
    alpha = 4.0
    gamma = 8.0

    # FFC notation
    # a = dot(grad(v), grad(u))*dx \
    #    - dot(avg(grad(v)), jump(u, n))*dS \
    #    - dot(jump(v, n), avg(grad(u)))*dS \
    #    + alpha/h('+')*dot(jump(v, n), jump(u, n))*dS \
    #    - dot(grad(v), mult(u,n))*ds \
    #    - dot(mult(v,n), grad(u))*ds \
    #    + gamma/h*v*u*ds

    a = inner(grad(v), grad(u)) * dx
    a -= inner(avg(grad(v)), jump(u, n)) * dS
    a -= inner(jump(v, n), avg(grad(u))) * dS
    a += alpha / h("+") * dot(jump(v, n), jump(u, n)) * dS
    a -= inner(grad(v), u * n) * ds
    a -= inner(u * n, grad(u)) * ds
    a += gamma / h * v * u * ds

    _ = v * f * dx + v * gN * ds


def testPoisson():
    element = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)

    v = TestFunction(space)
    u = TrialFunction(space)
    f = Coefficient(space)

    # Note: inner() also works
    _ = dot(grad(v), grad(u)) * dx
    _ = v * f * dx


def testPoissonSystem():
    element = FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)

    v = TestFunction(space)
    u = TrialFunction(space)
    f = Coefficient(space)

    # FFC notation: a = dot(grad(v), grad(u))*dx
    _ = inner(grad(v), grad(u)) * dx

    # FFC notation: L = dot(v, f)*dx
    _ = inner(v, f) * dx


def testProjection():
    # Projections are not supported by UFL and have been broken
    # in FFC for a while. For DOLFIN, the current (global) L^2
    # projection can be extended to handle also local projections.

    P1 = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, P1)
    _ = TestFunction(space)
    _ = Coefficient(space)

    # pi0 = Projection(P0)
    # pi1 = Projection(P1)
    # pi2 = Projection(P2)
    #
    # a = v*(pi0(f) + pi1(f) + pi2(f))*dx


def testQuadratureElement():
    element = FiniteElement("Lagrange", triangle, 2, (), identity_pullback, H1)

    # FFC notation:
    # QE = QuadratureElement(triangle, 3)
    # sig = VectorQuadratureElement(triangle, 3)

    QE = FiniteElement("Quadrature", triangle, 3, (), identity_pullback, L2)
    sig = FiniteElement("Quadrature", triangle, 3, (2,), identity_pullback, L2)

    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)

    v = TestFunction(space)
    u = TrialFunction(space)
    u0 = Coefficient(space)
    C = Coefficient(FunctionSpace(domain, QE))
    sig0 = Coefficient(FunctionSpace(domain, sig))
    f = Coefficient(space)

    _ = v.dx(i) * C * u.dx(i) * dx + v.dx(i) * 2 * u0 * u * u0.dx(i) * dx
    _ = v * f * dx - dot(grad(v), sig0) * dx


def testStokes():
    # UFLException: Shape mismatch in sum.

    P2 = FiniteElement("Lagrange", triangle, 2, (2,), identity_pullback, H1)
    P1 = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    TH = MixedElement([P2, P1])

    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    th_space = FunctionSpace(domain, TH)
    p2_space = FunctionSpace(domain, P2)

    (v, q) = TestFunctions(th_space)
    (u, p) = TrialFunctions(th_space)
    f = Coefficient(p2_space)

    # FFC notation:
    # a = (dot(grad(v), grad(u)) - div(v)*p + q*div(u))*dx
    _ = (inner(grad(v), grad(u)) - div(v) * p + q * div(u)) * dx
    _ = dot(v, f) * dx


def testSubDomain():
    element = FiniteElement("Lagrange", tetrahedron, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", tetrahedron, 1, (3,), identity_pullback, H1))
    space = FunctionSpace(domain, element)
    f = Coefficient(space)
    _ = f * dx(2) + f * ds(5)


def testSubDomains():
    element = FiniteElement("Lagrange", tetrahedron, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", tetrahedron, 1, (3,), identity_pullback, H1))
    space = FunctionSpace(domain, element)

    v = TestFunction(space)
    u = TrialFunction(space)
    a = v * u * dx(0) + 10.0 * v * u * dx(1) + v * u * ds(0) + 2.0 * v * u * ds(1)
    a += v("+") * u("+") * dS(0) + 4.3 * v("+") * u("+") * dS(1)


def testTensorWeightedPoisson():
    # FFC notation:
    # P1 = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    # P0 = FiniteElement("Discontinuous Lagrange", triangle, 0, (), identity_pullback, L2)
    #
    # v = TestFunction(P1)
    # u = TrialFunction(P1)
    # f = Coefficient(P1)
    #
    # c00 = Coefficient(P0)
    # c01 = Coefficient(P0)
    # c10 = Coefficient(P0)
    # c11 = Coefficient(P0)
    #
    # C = [[c00, c01], [c10, c11]]
    #
    # a = dot(grad(v), mult(C, grad(u)))*dx

    P1 = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    P0 = FiniteElement("Discontinuous Lagrange", triangle, 0, (2, 2), identity_pullback, L2)

    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    p1_space = FunctionSpace(domain, P1)
    p0_space = FunctionSpace(domain, P0)

    v = TestFunction(p1_space)
    u = TrialFunction(p1_space)
    C = Coefficient(p0_space)
    _ = inner(grad(v), C * grad(u)) * dx


def testVectorLaplaceGradCurl():
    def HodgeLaplaceGradCurl(space, fspace):
        (tau, v) = TestFunctions(space)
        (sigma, u) = TrialFunctions(space)
        f = Coefficient(fspace)

        # FFC notation: a = (dot(tau, sigma) - dot(grad(tau), u) + dot(v,
        # grad(sigma)) + dot(curl(v), curl(u)))*dx
        a = (
            inner(tau, sigma)
            - inner(grad(tau), u)
            + inner(v, grad(sigma))
            + inner(curl(v), curl(u))
        ) * dx

        # FFC notation: L = dot(v, f)*dx
        L = inner(v, f) * dx

        return [a, L]

    shape = tetrahedron
    order = 1

    GRAD = FiniteElement("Lagrange", shape, order, (), identity_pullback, H1)

    # FFC notation: CURL = FiniteElement("Nedelec", shape, order-1)
    CURL = FiniteElement("N1curl", shape, order, (3,), covariant_piola, HCurl)

    VectorLagrange = FiniteElement("Lagrange", shape, order + 1, (3,), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", shape, 1, (3,), identity_pullback, H1))

    [a, L] = HodgeLaplaceGradCurl(
        FunctionSpace(domain, MixedElement([GRAD, CURL])), FunctionSpace(domain, VectorLagrange)
    )



"""Types for representing a cell."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from __future__ import annotations

import functools
import numbers
import typing
import weakref
from abc import abstractmethod

from ufl.core.ufl_type import UFLObject

__all_classes__ = ["AbstractCell", "Cell", "TensorProductCell"]


class AbstractCell(UFLObject):
    """A base class for all cells."""

    @abstractmethod
    def topological_dimension(self) -> int:
        """Return the dimension of the topology of this cell."""

    @abstractmethod
    def is_simplex(self) -> bool:
        """Return True if this is a simplex cell."""

    @abstractmethod
    def has_simplex_facets(self) -> bool:
        """Return True if all the facets of this cell are simplex cells."""

    @abstractmethod
    def _lt(self, other) -> bool:
        """Less than operator.

        Define an arbitrarily chosen but fixed sort order for all
        instances of this type with the same dimensions.
        """

    @abstractmethod
    def num_sub_entities(self, dim: int) -> int:
        """Get the number of sub-entities of the given dimension."""

    @abstractmethod
    def sub_entities(self, dim: int) -> typing.Tuple[AbstractCell, ...]:
        """Get the sub-entities of the given dimension."""

    @abstractmethod
    def sub_entity_types(self, dim: int) -> typing.Tuple[AbstractCell, ...]:
        """Get the unique sub-entity types of the given dimension."""

    @abstractmethod
    def cellname(self) -> str:
        """Return the cellname of the cell."""

    @abstractmethod
    def reconstruct(self, **kwargs: typing.Any) -> Cell:
        """Reconstruct this cell, overwriting properties by those in kwargs."""

    def __lt__(self, other: AbstractCell) -> bool:
        """Define an arbitrarily chosen but fixed sort order for all cells."""
        if type(self) is type(other):
            s = self.topological_dimension()
            o = other.topological_dimension()
            if s != o:
                return s < o
            return self._lt(other)
        else:
            if type(self).__name__ == type(other).__name__:
                raise ValueError("Cannot order cell types with the same name")
            return type(self).__name__ < type(other).__name__

    def num_vertices(self) -> int:
        """Get the number of vertices."""
        return self.num_sub_entities(0)

    def num_edges(self) -> int:
        """Get the number of edges."""
        return self.num_sub_entities(1)

    def num_faces(self) -> int:
        """Get the number of faces."""
        return self.num_sub_entities(2)

    def num_facets(self) -> int:
        """Get the number of facets.

        Facets are entities of dimension tdim-1.
        """
        tdim = self.topological_dimension()
        return self.num_sub_entities(tdim - 1)

    def num_ridges(self) -> int:
        """Get the number of ridges.

        Ridges are entities of dimension tdim-2.
        """
        tdim = self.topological_dimension()
        return self.num_sub_entities(tdim - 2)

    def num_peaks(self) -> int:
        """Get the number of peaks.

        Peaks are entities of dimension tdim-3.
        """
        tdim = self.topological_dimension()
        return self.num_sub_entities(tdim - 3)

    def vertices(self) -> typing.Tuple[AbstractCell, ...]:
        """Get the vertices."""
        return self.sub_entities(0)

    def edges(self) -> typing.Tuple[AbstractCell, ...]:
        """Get the edges."""
        return self.sub_entities(1)

    def faces(self) -> typing.Tuple[AbstractCell, ...]:
        """Get the faces."""
        return self.sub_entities(2)

    def facets(self) -> typing.Tuple[AbstractCell, ...]:
        """Get the facets.

        Facets are entities of dimension tdim-1.
        """
        tdim = self.topological_dimension()
        return self.sub_entities(tdim - 1)

    def ridges(self) -> typing.Tuple[AbstractCell, ...]:
        """Get the ridges.

        Ridges are entities of dimension tdim-2.
        """
        tdim = self.topological_dimension()
        return self.sub_entities(tdim - 2)

    def peaks(self) -> typing.Tuple[AbstractCell, ...]:
        """Get the peaks.

        Peaks are entities of dimension tdim-3.
        """
        tdim = self.topological_dimension()
        return self.sub_entities(tdim - 3)

    def vertex_types(self) -> typing.Tuple[AbstractCell, ...]:
        """Get the unique vertices types."""
        return self.sub_entity_types(0)

    def edge_types(self) -> typing.Tuple[AbstractCell, ...]:
        """Get the unique edge types."""
        return self.sub_entity_types(1)

    def face_types(self) -> typing.Tuple[AbstractCell, ...]:
        """Get the unique face types."""
        return self.sub_entity_types(2)

    def facet_types(self) -> typing.Tuple[AbstractCell, ...]:
        """Get the unique facet types.

        Facets are entities of dimension tdim-1.
        """
        tdim = self.topological_dimension()
        return self.sub_entity_types(tdim - 1)

    def ridge_types(self) -> typing.Tuple[AbstractCell, ...]:
        """Get the unique ridge types.

        Ridges are entities of dimension tdim-2.
        """
        tdim = self.topological_dimension()
        return self.sub_entity_types(tdim - 2)

    def peak_types(self) -> typing.Tuple[AbstractCell, ...]:
        """Get the unique peak types.

        Peaks are entities of dimension tdim-3.
        """
        tdim = self.topological_dimension()
        return self.sub_entity_types(tdim - 3)


_sub_entity_celltypes = {
    "vertex": [("vertex",)],
    "interval": [tuple("vertex" for i in range(2)), ("interval",)],
    "triangle": [
        tuple("vertex" for i in range(3)),
        tuple("interval" for i in range(3)),
        ("triangle",),
    ],
    "quadrilateral": [
        tuple("vertex" for i in range(4)),
        tuple("interval" for i in range(4)),
        ("quadrilateral",),
    ],
    "tetrahedron": [
        tuple("vertex" for i in range(4)),
        tuple("interval" for i in range(6)),
        tuple("triangle" for i in range(4)),
        ("tetrahedron",),
    ],
    "hexahedron": [
        tuple("vertex" for i in range(8)),
        tuple("interval" for i in range(12)),
        tuple("quadrilateral" for i in range(6)),
        ("hexahedron",),
    ],
    "prism": [
        tuple("vertex" for i in range(6)),
        tuple("interval" for i in range(9)),
        ("triangle", "quadrilateral", "quadrilateral", "quadrilateral", "triangle"),
        ("prism",),
    ],
    "pyramid": [
        tuple("vertex" for i in range(5)),
        tuple("interval" for i in range(8)),
        ("quadrilateral", "triangle", "triangle", "triangle", "triangle"),
        ("pyramid",),
    ],
    "pentatope": [
        tuple("vertex" for i in range(5)),
        tuple("interval" for i in range(10)),
        tuple("triangle" for i in range(10)),
        tuple("tetrahedron" for i in range(5)),
        ("pentatope",),
    ],
    "tesseract": [
        tuple("vertex" for i in range(16)),
        tuple("interval" for i in range(32)),
        tuple("quadrilateral" for i in range(24)),
        tuple("hexahedron" for i in range(8)),
        ("tesseract",),
    ],
}


class Cell(AbstractCell):
    """Representation of a named finite element cell with known structure."""

    __slots__ = (
        "_cellname",
        "_tdim",
        "_num_cell_entities",
        "_sub_entity_types",
        "_sub_entities",
        "_sub_entity_types",
    )

    def __init__(self, cellname: str):
        """Initialise.

        Args:
            cellname: Name of the cell
        """
        if cellname not in _sub_entity_celltypes:
            raise ValueError(f"Unsupported cell type: {cellname}")

        self._sub_entity_celltypes = _sub_entity_celltypes[cellname]

        self._cellname = cellname
        self._tdim = len(self._sub_entity_celltypes) - 1

        self._num_cell_entities = [len(i) for i in self._sub_entity_celltypes]
        self._sub_entities = [
            tuple(Cell(t) for t in se_types) for se_types in self._sub_entity_celltypes[:-1]
        ]
        self._sub_entity_types = [tuple(set(i)) for i in self._sub_entities]
        self._sub_entities.append((weakref.proxy(self),))
        self._sub_entity_types.append((weakref.proxy(self),))

        if not isinstance(self._tdim, numbers.Integral):
            raise ValueError("Expecting integer topological_dimension.")

    def topological_dimension(self) -> int:
        """Return the dimension of the topology of this cell."""
        return self._tdim

    def is_simplex(self) -> bool:
        """Return True if this is a simplex cell."""
        return self._cellname in ["vertex", "interval", "triangle", "tetrahedron"]

    def has_simplex_facets(self) -> bool:
        """Return True if all the facets of this cell are simplex cells."""
        return self._cellname in ["interval", "triangle", "quadrilateral", "tetrahedron"]

    def num_sub_entities(self, dim: int) -> int:
        """Get the number of sub-entities of the given dimension."""
        if dim < 0:
            return 0
        try:
            return self._num_cell_entities[dim]
        except IndexError:
            return 0

    def sub_entities(self, dim: int) -> typing.Tuple[AbstractCell, ...]:
        """Get the sub-entities of the given dimension."""
        if dim < 0:
            return ()
        try:
            return self._sub_entities[dim]
        except IndexError:
            return ()

    def sub_entity_types(self, dim: int) -> typing.Tuple[AbstractCell, ...]:
        """Get the unique sub-entity types of the given dimension."""
        if dim < 0:
            return ()
        try:
            return self._sub_entity_types[dim]
        except IndexError:
            return ()

    def _lt(self, other) -> bool:
        return self._cellname < other._cellname

    def cellname(self) -> str:
        """Return the cellname of the cell."""
        return self._cellname

    def __str__(self) -> str:
        """Format as a string."""
        return self._cellname

    def __repr__(self) -> str:
        """Representation."""
        return self._cellname

    def _ufl_hash_data_(self) -> typing.Hashable:
        """UFL hash data."""
        return (self._cellname,)

    def reconstruct(self, **kwargs: typing.Any) -> Cell:
        """Reconstruct this cell, overwriting properties by those in kwargs."""
        for key, value in kwargs.items():
            raise TypeError(f"reconstruct() got unexpected keyword argument '{key}'")
        return Cell(self._cellname)


class TensorProductCell(AbstractCell):
    """Tensor product cell."""

    __slots__ = ("_cells", "_tdim")

    def __init__(self, *cells: Cell):
        """Initialise.

        Args:
            cells: Cells to take the tensor product of
        """
        self._cells = tuple(as_cell(cell) for cell in cells)

        self._tdim = sum([cell.topological_dimension() for cell in self._cells])

        if not isinstance(self._tdim, numbers.Integral):
            raise ValueError("Expecting integer topological_dimension.")

    def sub_cells(self) -> typing.List[AbstractCell]:
        """Return list of cell factors."""
        return self._cells

    def topological_dimension(self) -> int:
        """Return the dimension of the topology of this cell."""
        return self._tdim

    def is_simplex(self) -> bool:
        """Return True if this is a simplex cell."""
        if len(self._cells) == 1:
            return self._cells[0].is_simplex()
        return False

    def has_simplex_facets(self) -> bool:
        """Return True if all the facets of this cell are simplex cells."""
        if len(self._cells) == 1:
            return self._cells[0].has_simplex_facets()
        if self._tdim == 1:
            return True
        return False

    def num_sub_entities(self, dim: int) -> int:
        """Get the number of sub-entities of the given dimension."""
        if dim < 0 or dim > self._tdim:
            return 0
        if dim == 0:
            return functools.reduce(lambda x, y: x * y, [c.num_vertices() for c in self._cells])
        if dim == self._tdim - 1:
            # Note: This is not the number of facets that the cell has,
            # but I'm leaving it here for now to not change past
            # behaviour
            return sum(c.num_facets() for c in self._cells if c.topological_dimension() > 0)
        if dim == self._tdim:
            return 1
        raise NotImplementedError(f"TensorProductCell.num_sub_entities({dim}) is not implemented.")

    def sub_entities(self, dim: int) -> typing.Tuple[AbstractCell, ...]:
        """Get the sub-entities of the given dimension."""
        if dim < 0 or dim > self._tdim:
            return []
        if dim == 0:
            return [Cell("vertex") for i in range(self.num_sub_entities(0))]
        if dim == self._tdim:
            return [self]
        raise NotImplementedError(f"TensorProductCell.sub_entities({dim}) is not implemented.")

    def sub_entity_types(self, dim: int) -> typing.Tuple[AbstractCell, ...]:
        """Get the unique sub-entity types of the given dimension."""
        if dim < 0 or dim > self._tdim:
            return []
        if dim == 0:
            return [Cell("vertex")]
        if dim == self._tdim:
            return [self]
        raise NotImplementedError(f"TensorProductCell.sub_entities({dim}) is not implemented.")

    def _lt(self, other) -> bool:
        return self._ufl_hash_data_() < other._ufl_hash_data_()

    def cellname(self) -> str:
        """Return the cellname of the cell."""
        return " * ".join([cell.cellname() for cell in self._cells])

    def __str__(self) -> str:
        """Format as a string."""
        return "TensorProductCell(" + ", ".join(f"{c!r}" for c in self._cells) + ")"

    def __repr__(self) -> str:
        """Representation."""
        return str(self)

    def _ufl_hash_data_(self) -> typing.Hashable:
        """UFL hash data."""
        return tuple(c._ufl_hash_data_() for c in self._cells)

    def reconstruct(self, **kwargs: typing.Any) -> Cell:
        """Reconstruct this cell, overwriting properties by those in kwargs."""
        for key, value in kwargs.items():
            raise TypeError(f"reconstruct() got unexpected keyword argument '{key}'")
        return TensorProductCell(*self._cells)


def simplex(topological_dimension: int):
    """Return a simplex cell of the given dimension."""
    if topological_dimension == 0:
        return Cell("vertex")
    if topological_dimension == 1:
        return Cell("interval")
    if topological_dimension == 2:
        return Cell("triangle")
    if topological_dimension == 3:
        return Cell("tetrahedron")
    if topological_dimension == 4:
        return Cell("pentatope")
    raise ValueError(f"Unsupported topological dimension for simplex: {topological_dimension}")


def hypercube(topological_dimension: int):
    """Return a hypercube cell of the given dimension."""
    if topological_dimension == 0:
        return Cell("vertex")
    if topological_dimension == 1:
        return Cell("interval")
    if topological_dimension == 2:
        return Cell("quadrilateral")
    if topological_dimension == 3:
        return Cell("hexahedron")
    if topological_dimension == 4:
        return Cell("tesseract")
    raise ValueError(f"Unsupported topological dimension for hypercube: {topological_dimension}")


def as_cell(cell: typing.Union[AbstractCell, str, typing.Tuple[AbstractCell, ...]]) -> AbstractCell:
    """Convert any valid object to a Cell or return cell if it is already a Cell.

    Allows an already valid cell, a known cellname string, or a tuple of cells for a product cell.
    """
    if isinstance(cell, AbstractCell):
        return cell
    elif isinstance(cell, str):
        return Cell(cell)
    elif isinstance(cell, tuple):
        return TensorProductCell(cell)
    else:
        raise ValueError(f"Invalid cell {cell}.")



"""Expr equals."""

from collections import defaultdict

hash_total = defaultdict(int)
hash_collisions = defaultdict(int)
hash_equals = defaultdict(int)
hash_notequals = defaultdict(int)


def expr_equals(self, other):
    """Checks whether the two expressions are represented the exact same way.

    This does not check if the expressions are
    mathematically equal or equivalent! Used by sets and dicts.
    """
    # Fast cutoffs for common cases, type difference or hash
    # difference will cutoff more or less all nonequal types
    if type(self) is not type(other) or hash(self) != hash(other):
        return False

    # Large objects are costly to compare with themselves
    if self is other:
        return True

    # Modelled after pre_traversal to avoid recursion:
    left = [(self, other)]
    while left:
        s, o = left.pop()

        if s._ufl_is_terminal_:
            # Compare terminals
            if not s == o:
                return False
        else:
            # Delve into subtrees
            so = s.ufl_operands
            oo = o.ufl_operands
            if len(so) != len(oo):
                return False

            for s, o in zip(so, oo):
                # Fast cutoff for common case
                if s._ufl_typecode_ != o._ufl_typecode_:
                    return False
                # Skip subtree if objects are the same
                if s is o:
                    continue
                # Append subtree for further inspection
                left.append((s, o))

    # Equal if we get out of the above loop!
    # Eagerly DAGify to reduce the size of the tree.
    self.ufl_operands = other.ufl_operands
    return True



__authors__ = "Marie E. Rognes"

# First added: 2011-11-09
# Last changed: 2011-11-09

from ufl import (
    Argument,
    Coefficient,
    Constant,
    FunctionSpace,
    Mesh,
    TestFunction,
    TrialFunction,
    action,
    derivative,
    dS,
    ds,
    dx,
    exp,
    interval,
    system,
)
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1


def test_lhs_rhs_simple():
    V = FiniteElement("Lagrange", interval, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", interval, 1, (1,), identity_pullback, H1))
    space = FunctionSpace(domain, V)
    v = TestFunction(space)
    u = TrialFunction(space)
    w = Argument(space, 2)  # This was 0, not sure why
    f = Coefficient(space)

    F0 = f * u * v * w * dx
    a, L = system(F0)
    assert len(a.integrals()) == 0
    assert len(L.integrals()) == 0

    F1 = derivative(F0, f)
    a, L = system(F1)
    assert len(a.integrals()) == 0
    assert len(L.integrals()) == 0

    F2 = action(F0, f)
    a, L = system(F2)
    assert len(a.integrals()) == 1
    assert len(L.integrals()) == 0

    F3 = action(F2, f)
    a, L = system(F3)
    assert len(L.integrals()) == 1


def test_lhs_rhs_derivatives():
    V = FiniteElement("Lagrange", interval, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", interval, 1, (1,), identity_pullback, H1))
    space = FunctionSpace(domain, V)
    v = TestFunction(space)
    u = TrialFunction(space)
    f = Coefficient(space)

    F0 = exp(f) * u * v * dx + v * dx + f * v * ds + exp(f)("+") * v * dS
    a, L = system(F0)
    assert len(a.integrals()) == 1
    assert len(L.integrals()) == 3

    derivative(F0, f)
    a, L = system(F0)


def test_lhs_rhs_slightly_obscure():
    V = FiniteElement("Lagrange", interval, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", interval, 1, (1,), identity_pullback, H1))
    space = FunctionSpace(domain, V)
    u = TrialFunction(space)
    w = Argument(space, 2)
    f = Constant(domain)

    # FIXME:
    # ufl.algorithsm.formtransformations.compute_form_with_arity
    # is not perfect, e.g. try
    # F = f*u*w*dx + f*w*dx
    F = f * u * w * dx
    a, L = system(F)
    assert len(a.integrals()) == 1
    assert len(L.integrals()) == 0

    F = f * w * dx
    a, L = system(F)
    assert len(L.integrals()) == 1



"""This module defines the Interpolate class."""

# Copyright (C) 2021 Nacime Bouziani
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Nacime Bouziani, 2021-2022

from ufl.action import Action
from ufl.argument import Argument, Coargument
from ufl.coefficient import Cofunction
from ufl.constantvalue import as_ufl
from ufl.core.base_form_operator import BaseFormOperator
from ufl.core.ufl_type import ufl_type
from ufl.duals import is_dual
from ufl.form import BaseForm, Form
from ufl.functionspace import AbstractFunctionSpace


@ufl_type(num_ops="varying", is_differential=True)
class Interpolate(BaseFormOperator):
    """Symbolic representation of the interpolation operator."""

    # Slots are disabled here because they cause trouble in PyDOLFIN
    # multiple inheritance pattern:
    _ufl_noslots_ = True

    def __init__(self, expr, v):
        """Initialise.

        Args:
            expr: a UFL expression to interpolate.
            v: the FunctionSpace to interpolate into or the Coargument
                defined on the dual of the FunctionSpace to interpolate into.
        """
        # This check could be more rigorous.
        dual_args = (Coargument, Cofunction, Form, Action, BaseFormOperator)

        if isinstance(v, AbstractFunctionSpace):
            if is_dual(v):
                raise ValueError("Expecting a primal function space.")
            v = Argument(v.dual(), 0)
        elif not isinstance(v, dual_args):
            raise ValueError(
                "Expecting the second argument to be FunctionSpace, FiniteElement or dual."
            )

        expr = as_ufl(expr)
        if isinstance(expr, dual_args):
            raise ValueError("Expecting the first argument to be primal.")

        # Reversed order convention
        argument_slots = (v, expr)
        # Get the primal space (V** = V)
        if isinstance(v, BaseForm):
            arg, *_ = v.arguments()
            function_space = arg.ufl_function_space()
        else:
            function_space = v.ufl_function_space().dual()
        # Set the operand as `expr` for DAG traversal purpose.
        operand = expr
        BaseFormOperator.__init__(
            self, operand, function_space=function_space, argument_slots=argument_slots
        )

    def _ufl_expr_reconstruct_(self, expr, v=None, **add_kwargs):
        """Return a new object of the same type with new operands."""
        v = v or self.argument_slots()[0]
        return type(self)(expr, v, **add_kwargs)

    def __repr__(self):
        """Default repr string construction for Interpolate."""
        r = "Interpolate("
        r += ", ".join(repr(arg) for arg in reversed(self.argument_slots()))
        r += f"; {self.ufl_function_space()!r})"
        return r

    def __str__(self):
        """Default str string construction for Interpolate."""
        s = "Interpolate("
        s += ", ".join(str(arg) for arg in reversed(self.argument_slots()))
        s += f"; {self.ufl_function_space()})"
        return s

    def __eq__(self, other):
        """Check for equality."""
        if self is other:
            return True
        return (
            type(self) is type(other)
            and all(a == b for a, b in zip(self._argument_slots, other._argument_slots))
            and self.ufl_function_space() == other.ufl_function_space()
        )


# Helper function
def interpolate(expr, v):
    """Create symbolic representation of the interpolation operator.

    Args:
        expr: a UFL expression to interpolate.
        v: the FunctionSpace to interpolate into or the Coargument
            defined on the dual of the FunctionSpace to interpolate into.
    """
    return Interpolate(expr, v)



# -*- coding: utf-8 -*-
# Copyright (C) 2004-2008 Anders Logg
#
# This file is part of UFL.
#
# UFL is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFL. If not, see <http://www.gnu.org/licenses/>.
#
# Modified by Martin Sandve Alnæs, 2009
#
# Last changed: 2009-03-02
#
# The bilinear form a(v, u) and linear form L(v) for Poisson's equation.
from ufl import (
    Coefficient,
    FunctionSpace,
    Mesh,
    TestFunction,
    TrialFunction,
    dx,
    grad,
    inner,
    triangle,
)
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1

element = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
space = FunctionSpace(domain, element)

u = TrialFunction(space)
v = TestFunction(space)
f = Coefficient(space)

a = inner(grad(v), grad(u)) * dx(degree=1)
L = v * f * dx(degree=2)



"""This module defines the Adjoint class."""

# Copyright (C) 2021 India Marsden
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Nacime Bouziani, 2021-2022.

from itertools import chain

from ufl.argument import Coargument
from ufl.core.ufl_type import ufl_type
from ufl.form import BaseForm, FormSum, ZeroBaseForm

# --- The Adjoint class represents the adjoint of a numerical object that
#     needs to be computed at assembly time ---


@ufl_type()
class Adjoint(BaseForm):
    """UFL base form type: represents the adjoint of an object.

    Adjoint objects will result when the adjoint of an assembled object
    (e.g. a Matrix) is taken. This delays the evaluation of the adjoint until
    assembly occurs.
    """

    __slots__ = (
        "_form",
        "_repr",
        "_arguments",
        "_coefficients",
        "_domains",
        "ufl_operands",
        "_hash",
    )

    def __new__(cls, *args, **kw):
        """Create a new Adjoint."""
        form = args[0]
        # Check trivial case: This is not a ufl.Zero but a ZeroBaseForm!
        if form == 0:
            # Swap the arguments
            return ZeroBaseForm(form.arguments()[::-1])

        if isinstance(form, Adjoint):
            return form._form
        elif isinstance(form, FormSum):
            # Adjoint distributes over sums
            return FormSum(*[(Adjoint(component), 1) for component in form.components()])
        elif isinstance(form, Coargument):
            # The adjoint of a coargument `c: V* -> V*` is the identity
            # matrix mapping from V to V (i.e. V x V* -> R).
            # Equivalently, the adjoint of `c` is its first argument,
            # which is a ufl.Argument defined on the primal space of
            # `c`.
            primal_arg, _ = form.arguments()
            # Returning the primal argument avoids explicit argument
            # reconstruction, making it a robust strategy for handling
            # subclasses of `ufl.Coargument`.
            return primal_arg

        return super(Adjoint, cls).__new__(cls)

    def __init__(self, form):
        """Initialise."""
        BaseForm.__init__(self)

        if len(form.arguments()) != 2:
            raise ValueError("Can only take Adjoint of a 2-form.")

        self._form = form
        self.ufl_operands = (self._form,)
        self._domains = None
        self._hash = None
        self._repr = "Adjoint(%s)" % repr(self._form)

    def ufl_function_spaces(self):
        """Get the tuple of function spaces of the underlying form."""
        return self._form.ufl_function_spaces()

    def form(self):
        """Return the form."""
        return self._form

    def _analyze_form_arguments(self):
        """The arguments of adjoint are the reverse of the form arguments."""
        reversed_args = self._form.arguments()[::-1]
        # Canonical numbering for arguments that is consistent with other BaseForm objects.
        self._arguments = tuple(
            type(arg)(arg.ufl_function_space(), number=i) for i, arg in enumerate(reversed_args)
        )
        self._coefficients = self._form.coefficients()

    def _analyze_domains(self):
        """Analyze which domains can be found in Adjoint."""
        from ufl.domain import join_domains

        # Collect unique domains
        self._domains = join_domains(
            chain.from_iterable(e.ufl_domains() for e in self.ufl_operands)
        )

    def ufl_domains(self):
        """Return all domains found in the base form."""
        if self._domains is None:
            self._analyze_domains()
        return self._domains

    def equals(self, other):
        """Check if two Adjoints are equal."""
        if type(other) is not Adjoint:
            return False
        if self is other:
            return True
        # Make sure we are returning a boolean as the equality can
        # result in a `ufl.Equation` if the underlying objects are
        # `ufl.BaseForm`.
        return bool(self._form == other._form)

    def __str__(self):
        """Format as a string."""
        return f"Adjoint({self._form})"

    def __repr__(self):
        """Representation."""
        return self._repr

    def __hash__(self):
        """Hash."""
        if self._hash is None:
            self._hash = hash(("Adjoint", hash(self._form)))
        return self._hash



# Copyright (C) 2005-2007 Anders Logg
#
# This file is part of UFL.
#
# UFL is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFL. If not, see <http://www.gnu.org/licenses/>.
#
# This example demonstrates how to define a functional, here
# the energy norm (squared) for a reaction-diffusion problem.
from ufl import Coefficient, FunctionSpace, Mesh, dot, dx, grad, tetrahedron
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1

element = FiniteElement("Lagrange", tetrahedron, 1, (), identity_pullback, H1)
domain = Mesh(FiniteElement("Lagrange", tetrahedron, 1, (3,), identity_pullback, H1))
space = FunctionSpace(domain, element)

v = Coefficient(space)
a = (v * v + dot(grad(v), grad(v))) * dx



"""Tests of the change to reference frame algorithm."""

from ufl import Coefficient, FunctionSpace, Mesh, triangle
from ufl.classes import Expr, ReferenceValue
from ufl.finiteelement import FiniteElement
from ufl.pullback import contravariant_piola, identity_pullback
from ufl.sobolevspace import H1, HDiv


def change_to_reference_frame(expr):
    assert isinstance(expr, Expr)
    return ReferenceValue(expr)


def test_change_unmapped_form_arguments_to_reference_frame():
    U = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    V = FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)
    T = FiniteElement("Lagrange", triangle, 1, (2, 2), identity_pullback, H1)

    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    u_space = FunctionSpace(domain, U)
    v_space = FunctionSpace(domain, V)
    t_space = FunctionSpace(domain, T)

    expr = Coefficient(u_space)
    assert change_to_reference_frame(expr) == ReferenceValue(expr)
    expr = Coefficient(v_space)
    assert change_to_reference_frame(expr) == ReferenceValue(expr)
    expr = Coefficient(t_space)
    assert change_to_reference_frame(expr) == ReferenceValue(expr)


def test_change_hdiv_form_arguments_to_reference_frame():
    V = FiniteElement("Raviart-Thomas", triangle, 1, (2,), contravariant_piola, HDiv)

    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    v_space = FunctionSpace(domain, V)

    expr = Coefficient(v_space)
    assert change_to_reference_frame(expr) == ReferenceValue(expr)


def test_change_hcurl_form_arguments_to_reference_frame():
    V = FiniteElement("Raviart-Thomas", triangle, 1, (2,), contravariant_piola, HDiv)

    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    v_space = FunctionSpace(domain, V)

    expr = Coefficient(v_space)
    assert change_to_reference_frame(expr) == ReferenceValue(expr)

    """
    # user input
    grad(f + g)('+')
    # change to reference frame
    -> (K*rgrad(rv(M)*rv(f) + rv(M)*rv(g)))('+')
    # apply derivatives
    -> (K*(rv(M)*rgrad(rv(f)) + rv(M)*rgrad(rv(g))))('+')
    # apply restrictions
    -> K('+')*(rv(M('+'))*rgrad(rv(f('+'))) + rv(M('+'))*rgrad(rv(g('+'))))



    # user input
    grad(f + g)('+')

    # some derivatives applied before processing
    (grad(f) + grad(g))('+')

    # ... replace to get fully defined form arguments here ...

    # expand compounds
    # * context options:
    #   - keep {types} without rewriting to lower level types
    #   - preserve div and curl if applied directly to terminal
    #     (ffc context may set this to off)
    # * output invariants:
    #   - no compound operator types left in expression (simplified language)
    #   - div and curl rewritten in terms of grad (optionally unless applied directly to terminal)
    -> (grad(f) + grad(g))('+')

    # change to reference frame
    # * context options:
    #   - keep {types} without rewriting to lower level types (e.g. JacobianInverse)
    #     (ffc context may initially add all code snippets expressions)
    #   - keep {types} in global frame (e.g. Coefficient)
    #     (ffc context may initially add Coefficient and Argument here to refrain from changing)
    #   - skip integral scaling
    #     (ffc context may turn skipping on to preserve current behaviour)
    # * output invariants:
    #   - ReferenceValue bound directly to terminals where applicable
    #   - grad replaced by mapping expression of rgrad
    #   - div replaced by mapping expression of rdiv
    #   - curl replaced by mapping expression of rcurl
    -> as_tensor(IndexSum(K[i,j]*rgrad(as_tensor(rv(M)[k,l]*rv(f)[l], (l,))
                                     + as_tensor(rv(M)[r,s]*rv(g)[s], (s,)))[j],
                          j),
                 (i,))('+')

    # apply derivatives
    # * context options:
    #   - N/A?
    # * output invariants:
    #   - grad,div,curl, bound directly to terminals
    #   - rgrad,rdiv,rcurl bound directly to referencevalue objects (rgrad(global_f) invalid)
    -> (K*(rv(M)*rgrad(rv(f)) + rv(M)*rgrad(rv(g))))('+')

    # apply restrictions
    # * context options:
    #   - N/A?
    # * output invariants:
    #   - *_restricted bound directly to terminals
    #   - all terminals that must be restricted to make sense are restricted
    -> K('+')*(rv(M('+'))*rgrad(rv(f('+'))) + rv(M('+'))*rgrad(rv(g('+'))))

    # final modified terminal structure:
    t = terminal | restricted(terminal)  # choice of terminal
    r = rval(t) | rgrad(r)               # in reference frame: value or n-gradient
    g = t | grad(g)                      # in global frame: value or n-gradient
    v = r | g                            # value in either frame
    e = v | cell_avg(v) | facet_avg(v) | at_cell_midpoint(v) | at_facet_midpoint(v)
                                         # evaluated at point or averaged over cell entity
    m = e | indexed(e)                   # scalar component of
    """


"""
New form preprocessing pipeline:

Preferably introduce these changes:
1) Create new FormArgument Expression without element or domain
2) Create new FormArgument Constant without domain
3) Drop replace
--> but just applying replace first is fine

i) group and join integrals by (domain, type, subdomain_id),
ii) process integrands:
    a) apply_coefficient_completion # replace coefficients to ensure proper elements and domains
    b) lower_compound_operators # expand_compounds
    c) change_to_reference_frame # change f->rv(f), m->M*rv(m),
                                          grad(f)->K*rgrad(rv(f)),
                                          grad(grad(f))->K*rgrad(K*rgrad(rv(f))),
                                                grad(expr)->K*rgrad(expr)
                                 # if grad(expr)->K*rgrad(expr) should be valid,
                                   then rgrad must be applicable to quite generic expressions
    d) apply_derivatives         # one possibility is to add an apply_mapped_derivatives AD
                                   algorithm which includes mappings
    e) apply_geometry_lowering
    f) apply_restrictions # requiring grad(f)('+') instead of grad(f('+')) would simplify a lot...
iii) extract final metadata about elements and coefficient ordering
"""



"""Types for representing function spaces."""

# Copyright (C) 2015-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Massimiliano Leoni, 2016
# Modified by Cecile Daversin-Catty, 2018

import typing

import numpy as np

from ufl.core.ufl_type import UFLObject
from ufl.domain import join_domains
from ufl.duals import is_dual, is_primal
from ufl.utils.sequences import product

# Export list for ufl.classes
__all_classes__ = [
    "AbstractFunctionSpace",
    "FunctionSpace",
    "DualSpace",
    "MixedFunctionSpace",
    "TensorProductFunctionSpace",
]


class AbstractFunctionSpace(object):
    """Abstract function space."""

    def ufl_sub_spaces(self):
        """Return ufl sub spaces."""
        raise NotImplementedError(
            f"Missing implementation of ufl_sub_spaces in {self.__class__.__name__}."
        )


class BaseFunctionSpace(AbstractFunctionSpace, UFLObject):
    """Base function space."""

    def __init__(self, domain, element, label=""):
        """Initialise."""
        if domain is None:
            # DOLFIN hack
            # TODO: Is anything expected from element.cell in this case?
            pass
        else:
            try:
                domain_cell = domain.ufl_cell()
            except AttributeError:
                raise ValueError(
                    "Expected non-abstract domain for initalization of function space."
                )
            else:
                if element.cell != domain_cell:
                    raise ValueError("Non-matching cell of finite element and domain.")
        AbstractFunctionSpace.__init__(self)
        self._label = label
        self._ufl_domain = domain
        self._ufl_element = element

    @property
    def components(self) -> typing.Dict[typing.Tuple[int, ...], int]:
        """Get the numbering of the components of the element of this space.

        Returns:
            A map from the components of the values on a physical cell (eg (0, 1))
            to flat component numbers on the reference cell (eg 1)
        """
        from ufl.pullback import SymmetricPullback

        if isinstance(self._ufl_element.pullback, SymmetricPullback):
            return self._ufl_element.pullback._symmetry

        if len(self._ufl_element.sub_elements) == 0:
            return {(): 0}

        components = {}
        offset = 0
        c_offset = 0
        for s in self.ufl_sub_spaces():
            for i, j in enumerate(np.ndindex(s.value_shape)):
                components[(offset + i,)] = c_offset + s.components[j]
            c_offset += max(s.components.values()) + 1
            offset += s.value_size
        return components

    def label(self):
        """Return label of boundary domains to differentiate restricted and unrestricted."""
        return self._label

    def ufl_sub_spaces(self):
        """Return ufl sub spaces."""
        return ()

    def ufl_domain(self):
        """Return ufl domain."""
        return self._ufl_domain

    def ufl_element(self):
        """Return ufl element."""
        return self._ufl_element

    def ufl_domains(self):
        """Return ufl domains."""
        domain = self.ufl_domain()
        if domain is None:
            return ()
        else:
            return (domain,)

    def _ufl_hash_data_(self, name=None):
        """UFL hash data."""
        name = name or "BaseFunctionSpace"
        domain = self.ufl_domain()
        element = self.ufl_element()
        if domain is None:
            ddata = None
        else:
            ddata = domain._ufl_hash_data_()
        if element is None:
            edata = None
        else:
            edata = element._ufl_hash_data_()
        return (name, ddata, edata, self.label())

    def _ufl_signature_data_(self, renumbering, name=None):
        """UFL signature data."""
        name = name or "BaseFunctionSpace"
        domain = self.ufl_domain()
        element = self.ufl_element()
        if domain is None:
            ddata = None
        else:
            ddata = domain._ufl_signature_data_(renumbering)
        if element is None:
            edata = None
        else:
            edata = element._ufl_signature_data_()
        return (name, ddata, edata, self.label())

    def __repr__(self):
        """Representation."""
        return f"BaseFunctionSpace({self._ufl_domain!r}, {self._ufl_element!r})"

    @property
    def value_shape(self) -> typing.Tuple[int, ...]:
        """Return the shape of the value space on a physical domain."""
        return self._ufl_element.pullback.physical_value_shape(self._ufl_element, self._ufl_domain)

    @property
    def value_size(self) -> int:
        """Return the integer product of the value shape on a physical domain."""
        return product(self.value_shape)


class FunctionSpace(BaseFunctionSpace, UFLObject):
    """Representation of a Function space."""

    _primal = True
    _dual = False

    def dual(self):
        """Get the dual of the space."""
        return DualSpace(self._ufl_domain, self._ufl_element, label=self.label())

    def _ufl_hash_data_(self):
        """UFL hash data."""
        return BaseFunctionSpace._ufl_hash_data_(self, "FunctionSpace")

    def _ufl_signature_data_(self, renumbering):
        """UFL signature data."""
        return BaseFunctionSpace._ufl_signature_data_(self, renumbering, "FunctionSpace")

    def __repr__(self):
        """Representation."""
        return f"FunctionSpace({self._ufl_domain!r}, {self._ufl_element!r})"

    def __str__(self):
        """String."""
        return f"FunctionSpace({self._ufl_domain}, {self._ufl_element})"


class DualSpace(BaseFunctionSpace, UFLObject):
    """Representation of a Dual space."""

    _primal = False
    _dual = True

    def __init__(self, domain, element, label=""):
        """Initialise."""
        BaseFunctionSpace.__init__(self, domain, element, label)

    def dual(self):
        """Get the dual of the space."""
        return FunctionSpace(self._ufl_domain, self._ufl_element, label=self.label())

    def _ufl_hash_data_(self):
        """UFL hash data."""
        return BaseFunctionSpace._ufl_hash_data_(self, "DualSpace")

    def _ufl_signature_data_(self, renumbering):
        """UFL signature data."""
        return BaseFunctionSpace._ufl_signature_data_(self, renumbering, "DualSpace")

    def __repr__(self):
        """Representation."""
        return f"DualSpace({self._ufl_domain!r}, {self._ufl_element!r})"

    def __str__(self):
        """String."""
        return f"DualSpace({self._ufl_domain}, {self._ufl_element})"


class TensorProductFunctionSpace(AbstractFunctionSpace, UFLObject):
    """Tensor product function space."""

    def __init__(self, *function_spaces):
        """Initialise."""
        AbstractFunctionSpace.__init__(self)
        self._ufl_function_spaces = function_spaces

    def ufl_sub_spaces(self):
        """Return ufl sub spaces."""
        return self._ufl_function_spaces

    def _ufl_hash_data_(self):
        """UFL hash data."""
        return ("TensorProductFunctionSpace",) + tuple(
            V._ufl_hash_data_() for V in self.ufl_sub_spaces()
        )

    def _ufl_signature_data_(self, renumbering):
        """UFL signature data."""
        return ("TensorProductFunctionSpace",) + tuple(
            V._ufl_signature_data_(renumbering) for V in self.ufl_sub_spaces()
        )

    def __repr__(self):
        """Representation."""
        return f"TensorProductFunctionSpace(*{self._ufl_function_spaces!r})"

    def __str__(self):
        """String."""
        return self.__repr__()


class MixedFunctionSpace(AbstractFunctionSpace, UFLObject):
    """Mixed function space."""

    def __init__(self, *args):
        """Initialise."""
        AbstractFunctionSpace.__init__(self)
        self._ufl_function_spaces = args
        self._ufl_elements = list()
        for fs in args:
            if isinstance(fs, BaseFunctionSpace):
                self._ufl_elements.append(fs.ufl_element())
            else:
                raise ValueError("Expecting BaseFunctionSpace objects")

        # A mixed FS is only primal/dual if all the subspaces are primal/dual"
        self._primal = all([is_primal(subspace) for subspace in self._ufl_function_spaces])
        self._dual = all([is_dual(subspace) for subspace in self._ufl_function_spaces])

    def ufl_sub_spaces(self):
        """Return ufl sub spaces."""
        return self._ufl_function_spaces

    def ufl_sub_space(self, i):
        """Return i-th ufl sub space."""
        return self._ufl_function_spaces[i]

    def dual(self, *args):
        """Return the dual to this function space.

        If no additional arguments are passed then a MixedFunctionSpace is
        returned whose components are the duals of the originals.

        If additional arguments are passed, these must be integers. In this
        case, the MixedFunctionSpace which is returned will have dual
        components in the positions corresponding to the arguments passed, and
        the original components in the other positions.
        """
        if args:
            spaces = [
                space.dual() if i in args else space
                for i, space in enumerate(self._ufl_function_spaces)
            ]
            return MixedFunctionSpace(*spaces)
        else:
            return MixedFunctionSpace(*[space.dual() for space in self._ufl_function_spaces])

    def ufl_elements(self):
        """Return ufl elements."""
        return self._ufl_elements

    def ufl_element(self):
        """Return ufl element."""
        if len(self._ufl_elements) == 1:
            return self._ufl_elements[0]
        else:
            raise ValueError(
                "Found multiple elements. Cannot return only one. "
                "Consider building a FunctionSpace from a MixedElement "
                "in case of homogeneous dimension."
            )

    def ufl_domains(self):
        """Return ufl domains."""
        domainlist = []
        for s in self._ufl_function_spaces:
            domainlist.extend(s.ufl_domains())
        return join_domains(domainlist)

    def ufl_domain(self):
        """Return ufl domain."""
        domains = self.ufl_domains()
        if len(domains) == 1:
            return domains[0]
        elif domains:
            raise ValueError("Found multiple domains, cannot return just one.")
        else:
            return None

    def num_sub_spaces(self):
        """Return number of subspaces."""
        return len(self._ufl_function_spaces)

    def _ufl_hash_data_(self):
        """UFL hash data."""
        return ("MixedFunctionSpace",) + tuple(V._ufl_hash_data_() for V in self.ufl_sub_spaces())

    def _ufl_signature_data_(self, renumbering):
        """UFL signature data."""
        return ("MixedFunctionSpace",) + tuple(
            V._ufl_signature_data_(renumbering) for V in self.ufl_sub_spaces()
        )

    def __repr__(self):
        """Representation."""
        return f"MixedFunctionSpace(*{self._ufl_function_spaces!r})"

    def __str__(self):
        """String."""
        return self.__repr__()



"""Pull back and push forward maps."""
# Copyright (C) 2023 Matthew Scroggs, David Ham, Garth Wells
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from __future__ import annotations

import typing
from abc import ABC, abstractmethod, abstractproperty
from typing import TYPE_CHECKING

import numpy as np

from ufl.core.expr import Expr
from ufl.core.multiindex import indices
from ufl.domain import extract_unique_domain
from ufl.functionspace import FunctionSpace
from ufl.tensors import as_tensor

if TYPE_CHECKING:
    from ufl.finiteelement import AbstractFiniteElement as _AbstractFiniteElement

__all_classes__ = [
    "NonStandardPullbackException",
    "AbstractPullback",
    "IdentityPullback",
    "ContravariantPiola",
    "CovariantPiola",
    "L2Piola",
    "DoubleContravariantPiola",
    "DoubleCovariantPiola",
    "MixedPullback",
    "SymmetricPullback",
    "PhysicalPullback",
    "CustomPullback",
    "UndefinedPullback",
]


class NonStandardPullbackException(BaseException):
    """Exception to raise if a map is non-standard."""

    pass


class AbstractPullback(ABC):
    """An abstract pull back."""

    @abstractmethod
    def __repr__(self) -> str:
        """Return a representation of the object."""

    @abstractmethod
    def physical_value_shape(self, element, domain) -> typing.Tuple[int, ...]:
        """Get the physical value shape when this pull back is applied to an element on a domain.

        Args:
            element: The element that the pull back is applied to
            domain: The domain

        Returns:
            The value shape when the pull back is applied to the given element
        """

    @abstractproperty
    def is_identity(self) -> bool:
        """Is this pull back the identity (or the identity applied to mutliple components)."""

    def apply(self, expr: Expr) -> Expr:
        """Apply the pull back.

        Args:
            expr: A function on a physical cell

        Returns: The function pulled back to the reference cell
        """
        raise NonStandardPullbackException()


class IdentityPullback(AbstractPullback):
    """The identity pull back."""

    def __repr__(self) -> str:
        """Return a representation of the object."""
        return "IdentityPullback()"

    @property
    def is_identity(self) -> bool:
        """Is this pull back the identity (or the identity applied to mutliple components)."""
        return True

    def apply(self, expr):
        """Apply the pull back.

        Args:
            expr: A function on a physical cell

        Returns: The function pulled back to the reference cell
        """
        return expr

    def physical_value_shape(self, element, domain) -> typing.Tuple[int, ...]:
        """Get the physical value shape when this pull back is applied to an element on a domain.

        Args:
            element: The element that the pull back is applied to
            domain: The domain

        Returns:
            The value shape when the pull back is applied to the given element
        """
        return element.reference_value_shape


class ContravariantPiola(AbstractPullback):
    """The contravariant Piola pull back."""

    def __repr__(self) -> str:
        """Return a representation of the object."""
        return "ContravariantPiola()"

    @property
    def is_identity(self) -> bool:
        """Is this pull back the identity (or the identity applied to mutliple components)."""
        return False

    def apply(self, expr):
        """Apply the pull back.

        Args:
            expr: A function on a physical cell

        Returns: The function pulled back to the reference cell
        """
        from ufl.classes import Jacobian, JacobianDeterminant

        domain = extract_unique_domain(expr)
        J = Jacobian(domain)
        detJ = JacobianDeterminant(J)
        transform = (1.0 / detJ) * J
        # Apply transform "row-wise" to TensorElement(PiolaMapped, ...)
        *k, i, j = indices(len(expr.ufl_shape) + 1)
        kj = (*k, j)
        return as_tensor(transform[i, j] * expr[kj], (*k, i))

    def physical_value_shape(self, element, domain) -> typing.Tuple[int, ...]:
        """Get the physical value shape when this pull back is applied to an element on a domain.

        Args:
            element: The element that the pull back is applied to
            domain: The domain

        Returns:
            The value shape when the pull back is applied to the given element
        """
        gdim = domain.geometric_dimension()
        return (gdim,) + element.reference_value_shape[1:]


class CovariantPiola(AbstractPullback):
    """The covariant Piola pull back."""

    def __repr__(self) -> str:
        """Return a representation of the object."""
        return "CovariantPiola()"

    @property
    def is_identity(self) -> bool:
        """Is this pull back the identity (or the identity applied to mutliple components)."""
        return False

    def apply(self, expr):
        """Apply the pull back.

        Args:
            expr: A function on a physical cell

        Returns: The function pulled back to the reference cell
        """
        from ufl.classes import JacobianInverse

        domain = extract_unique_domain(expr)
        K = JacobianInverse(domain)
        # Apply transform "row-wise" to TensorElement(PiolaMapped, ...)
        *k, i, j = indices(len(expr.ufl_shape) + 1)
        kj = (*k, j)
        return as_tensor(K[j, i] * expr[kj], (*k, i))

    def physical_value_shape(self, element, domain) -> typing.Tuple[int, ...]:
        """Get the physical value shape when this pull back is applied to an element on a domain.

        Args:
            element: The element that the pull back is applied to
            domain: The domain

        Returns:
            The value shape when the pull back is applied to the given element
        """
        gdim = domain.geometric_dimension()
        return (gdim,) + element.reference_value_shape[1:]


class L2Piola(AbstractPullback):
    """The L2 Piola pull back."""

    def __repr__(self) -> str:
        """Return a representation of the object."""
        return "L2Piola()"

    @property
    def is_identity(self) -> bool:
        """Is this pull back the identity (or the identity applied to mutliple components)."""
        return False

    def apply(self, expr):
        """Apply the pull back.

        Args:
            expr: A function on a physical cell

        Returns: The function pulled back to the reference cell
        """
        from ufl.classes import JacobianDeterminant

        domain = extract_unique_domain(expr)
        detJ = JacobianDeterminant(domain)
        return expr / detJ

    def physical_value_shape(self, element, domain) -> typing.Tuple[int, ...]:
        """Get the physical value shape when this pull back is applied to an element on a domain.

        Args:
            element: The element that the pull back is applied to
            domain: The domain

        Returns:
            The value shape when the pull back is applied to the given element
        """
        return element.reference_value_shape


class DoubleContravariantPiola(AbstractPullback):
    """The double contravariant Piola pull back."""

    def __repr__(self) -> str:
        """Return a representation of the object."""
        return "DoubleContravariantPiola()"

    @property
    def is_identity(self) -> bool:
        """Is this pull back the identity (or the identity applied to mutliple components)."""
        return False

    def apply(self, expr):
        """Apply the pull back.

        Args:
            expr: A function on a physical cell

        Returns: The function pulled back to the reference cell
        """
        from ufl.classes import Jacobian, JacobianDeterminant

        domain = extract_unique_domain(expr)
        J = Jacobian(domain)
        detJ = JacobianDeterminant(J)
        # Apply transform "row-wise" to TensorElement(PiolaMapped, ...)
        *k, i, j, m, n = indices(len(expr.ufl_shape) + 2)
        kmn = (*k, m, n)
        return as_tensor((1.0 / detJ) ** 2 * J[i, m] * expr[kmn] * J[j, n], (*k, i, j))

    def physical_value_shape(self, element, domain) -> typing.Tuple[int, ...]:
        """Get the physical value shape when this pull back is applied to an element on a domain.

        Args:
            element: The element that the pull back is applied to
            domain: The domain

        Returns:
            The value shape when the pull back is applied to the given element
        """
        gdim = domain.geometric_dimension()
        return (gdim, gdim)


class DoubleCovariantPiola(AbstractPullback):
    """The double covariant Piola pull back."""

    def __repr__(self) -> str:
        """Return a representation of the object."""
        return "DoubleCovariantPiola()"

    @property
    def is_identity(self) -> bool:
        """Is this pull back the identity (or the identity applied to mutliple components)."""
        return False

    def apply(self, expr):
        """Apply the pull back.

        Args:
            expr: A function on a physical cell

        Returns: The function pulled back to the reference cell
        """
        from ufl.classes import JacobianInverse

        domain = extract_unique_domain(expr)
        K = JacobianInverse(domain)
        # Apply transform "row-wise" to TensorElement(PiolaMapped, ...)
        *k, i, j, m, n = indices(len(expr.ufl_shape) + 2)
        kmn = (*k, m, n)
        return as_tensor(K[m, i] * expr[kmn] * K[n, j], (*k, i, j))

    def physical_value_shape(self, element, domain) -> typing.Tuple[int, ...]:
        """Get the physical value shape when this pull back is applied to an element on a domain.

        Args:
            element: The element that the pull back is applied to
            domain: The domain

        Returns:
            The value shape when the pull back is applied to the given element
        """
        gdim = domain.geometric_dimension()
        return (gdim, gdim)


class MixedPullback(AbstractPullback):
    """Pull back for a mixed element."""

    def __init__(self, element: _AbstractFiniteElement):
        """Initalise.

        Args:
            element: The mixed element
        """
        self._element = element

    def __repr__(self) -> str:
        """Return a representation of the object."""
        return f"MixedPullback({self._element!r})"

    @property
    def is_identity(self) -> bool:
        """Is this pull back the identity (or the identity applied to mutliple components)."""
        return all(e.pullback.is_identity for e in self._element.sub_elements)

    def apply(self, expr):
        """Apply the pull back.

        Args:
            expr: A function on a physical cell

        Returns: The function pulled back to the reference cell
        """
        domain = extract_unique_domain(expr)
        space = FunctionSpace(domain, self._element)
        rflat = [expr[idx] for idx in np.ndindex(expr.ufl_shape)]
        g_components = []
        offset = 0
        # For each unique piece in reference space, apply the appropriate pullback
        for subelem in self._element.sub_elements:
            rsub = as_tensor(
                np.asarray(rflat[offset : offset + subelem.reference_value_size]).reshape(
                    subelem.reference_value_shape
                )
            )
            rmapped = subelem.pullback.apply(rsub)
            # Flatten into the pulled back expression for the whole thing
            g_components.extend([rmapped[idx] for idx in np.ndindex(rmapped.ufl_shape)])
            offset += subelem.reference_value_size
        # And reshape appropriately
        f = as_tensor(np.asarray(g_components).reshape(space.value_shape))
        if f.ufl_shape != space.value_shape:
            raise ValueError(
                "Expecting pulled back expression with shape "
                f"'{space.value_shape}', got '{f.ufl_shape}'"
            )
        return f

    def physical_value_shape(self, element, domain) -> typing.Tuple[int, ...]:
        """Get the physical value shape when this pull back is applied to an element on a domain.

        Args:
            element: The element that the pull back is applied to
            domain: The domain

        Returns:
            The value shape when the pull back is applied to the given element
        """
        assert element == self._element
        dim = sum(FunctionSpace(domain, e).value_size for e in self._element.sub_elements)
        return (dim,)


class SymmetricPullback(AbstractPullback):
    """Pull back for an element with symmetry."""

    def __init__(
        self, element: _AbstractFiniteElement, symmetry: typing.Dict[typing.tuple[int, ...], int]
    ):
        """Initalise.

        Args:
            element: The element
            symmetry: A dictionary mapping from the component in
                physical space to the local component
        """
        self._element = element
        self._symmetry = symmetry

        self._sub_element_value_shape = element.sub_elements[0].reference_value_shape
        for e in element.sub_elements:
            if e.reference_value_shape != self._sub_element_value_shape:
                raise ValueError("Sub-elements must all have the same value shape.")
        self._block_shape = tuple(i + 1 for i in max(symmetry.keys()))

    def __repr__(self) -> str:
        """Return a representation of the object."""
        return f"SymmetricPullback({self._element!r}, {self._symmetry!r})"

    @property
    def is_identity(self) -> bool:
        """Is this pull back the identity (or the identity applied to mutliple components)."""
        return all(e.pullback.is_identity for e in self._element.sub_elements)

    def apply(self, expr):
        """Apply the pull back.

        Args:
            expr: A function on a physical cell

        Returns: The function pulled back to the reference cell
        """
        domain = extract_unique_domain(expr)
        space = FunctionSpace(domain, self._element)
        rflat = [expr[idx] for idx in np.ndindex(expr.ufl_shape)]
        g_components = []
        offsets = [0]
        for subelem in self._element.sub_elements:
            offsets.append(offsets[-1] + subelem.reference_value_size)
        # For each unique piece in reference space, apply the appropriate pullback
        for component in np.ndindex(self._block_shape):
            i = self._symmetry[component]
            subelem = self._element.sub_elements[i]
            rsub = as_tensor(
                np.asarray(rflat[offsets[i] : offsets[i + 1]]).reshape(
                    subelem.reference_value_shape
                )
            )
            rmapped = subelem.pullback.apply(rsub)
            # Flatten into the pulled back expression for the whole thing
            g_components.extend([rmapped[idx] for idx in np.ndindex(rmapped.ufl_shape)])
        # And reshape appropriately
        f = as_tensor(np.asarray(g_components).reshape(space.value_shape))
        if f.ufl_shape != space.value_shape:
            raise ValueError(
                f"Expecting pulled back expression with shape "
                f"'{space.value_shape}', got '{f.ufl_shape}'"
            )
        return f

    def physical_value_shape(self, element, domain) -> typing.Tuple[int, ...]:
        """Get the physical value shape when this pull back is applied to an element on a domain.

        Args:
            element: The element that the pull back is applied to
            domain: The domain

        Returns:
            The value shape when the pull back is applied to the given element
        """
        assert element == self._element
        return tuple(i + 1 for i in max(self._symmetry.keys()))


class PhysicalPullback(AbstractPullback):
    """Physical pull back.

    This should probably be removed.
    """

    def __repr__(self) -> str:
        """Return a representation of the object."""
        return "PhysicalPullback()"

    @property
    def is_identity(self) -> bool:
        """Is this pull back the identity (or the identity applied to mutliple components)."""
        return True

    def apply(self, expr):
        """Apply the pull back.

        Args:
            expr: A function on a physical cell

        Returns: The function pulled back to the reference cell
        """
        return expr

    def physical_value_shape(self, element, domain) -> typing.Tuple[int, ...]:
        """Get the physical value shape when this pull back is applied to an element on a domain.

        Args:
            element: The element that the pull back is applied to
            domain: The domain

        Returns:
            The value shape when the pull back is applied to the given element
        """
        raise NotImplementedError()


class CustomPullback(AbstractPullback):
    """Custom pull back.

    This should probably be removed.
    """

    def __repr__(self) -> str:
        """Return a representation of the object."""
        return "CustomPullback()"

    @property
    def is_identity(self) -> bool:
        """Is this pull back the identity (or the identity applied to mutliple components)."""
        return True

    def apply(self, expr):
        """Apply the pull back.

        Args:
            expr: A function on a physical cell

        Returns: The function pulled back to the reference cell
        """
        return expr

    def physical_value_shape(self, element, domain) -> typing.Tuple[int, ...]:
        """Get the physical value shape when this pull back is applied to an element on a domain.

        Args:
            element: The element that the pull back is applied to
            domain: The domain

        Returns:
            The value shape when the pull back is applied to the given element
        """
        if element.reference_value_shape == ():
            return ()
        raise NotImplementedError()


class UndefinedPullback(AbstractPullback):
    """Undefined pull back.

    This should probably be removed.
    """

    def __repr__(self) -> str:
        """Return a representation of the object."""
        return "UndefinedPullback()"

    @property
    def is_identity(self) -> bool:
        """Is this pull back the identity (or the identity applied to mutliple components)."""
        return True

    def physical_value_shape(self, element, domain) -> typing.Tuple[int, ...]:
        """Get the physical value shape when this pull back is applied to an element on a domain.

        Args:
            element: The element that the pull back is applied to
            domain: The domain

        Returns:
            The value shape when the pull back is applied to the given element
        """
        raise NotImplementedError()


identity_pullback = IdentityPullback()
covariant_piola = CovariantPiola()
contravariant_piola = ContravariantPiola()
l2_piola = L2Piola()
double_covariant_piola = DoubleCovariantPiola()
double_contravariant_piola = DoubleContravariantPiola()
physical_pullback = PhysicalPullback()
custom_pullback = CustomPullback()
undefined_pullback = UndefinedPullback()



"""This module defines classes for conditional expressions."""
# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

import warnings

from ufl.checks import is_true_ufl_scalar
from ufl.constantvalue import as_ufl
from ufl.core.expr import ufl_err_str
from ufl.core.operator import Operator
from ufl.core.ufl_type import ufl_type
from ufl.exprequals import expr_equals
from ufl.precedence import parstr

# --- Condition classes ---

# TODO: Would be nice with some kind of type system to show that this
# is a boolean type not a float type


@ufl_type(is_abstract=True, is_scalar=True)
class Condition(Operator):
    """Condition."""

    __slots__ = ()

    def __init__(self, operands):
        """Initialise."""
        Operator.__init__(self, operands)

    def __bool__(self):
        """Convert to a bool."""
        # Showing explicit error here to protect against misuse
        raise ValueError("UFL conditions cannot be evaluated as bool in a Python context.")

    __nonzero__ = __bool__


@ufl_type(is_abstract=True, num_ops=2)
class BinaryCondition(Condition):
    """Binary condition."""

    __slots__ = ("_name",)

    def __init__(self, name, left, right):
        """Initialise."""
        left = as_ufl(left)
        right = as_ufl(right)

        Condition.__init__(self, (left, right))

        self._name = name

        if name in ("!=", "=="):
            # Since equals and not-equals are used for comparing
            # representations, we have to allow any shape here. The
            # scalar properties must be checked when used in
            # conditional instead!
            pass
        elif name in ("&&", "||"):
            # Binary operators acting on boolean expressions allow
            # only conditions
            for arg in (left, right):
                if not isinstance(arg, Condition):
                    raise ValueError(f"Expecting a Condition, not {ufl_err_str(arg)}.")
        else:
            # Binary operators acting on non-boolean expressions allow
            # only scalars
            if left.ufl_shape != () or right.ufl_shape != ():
                raise ValueError("Expecting scalar arguments.")
            if left.ufl_free_indices != () or right.ufl_free_indices != ():
                raise ValueError("Expecting scalar arguments.")

    def __str__(self):
        """Format as a string."""
        return "%s %s %s" % (
            parstr(self.ufl_operands[0], self),
            self._name,
            parstr(self.ufl_operands[1], self),
        )


# Not associating with __eq__, the concept of equality with == is
# reserved for object equivalence for use in set and dict.
@ufl_type()
class EQ(BinaryCondition):
    """Equality condition."""

    __slots__ = ()

    def __init__(self, left, right):
        """Initialise."""
        BinaryCondition.__init__(self, "==", left, right)

    def evaluate(self, x, mapping, component, index_values):
        """Evaluate."""
        a = self.ufl_operands[0].evaluate(x, mapping, component, index_values)
        b = self.ufl_operands[1].evaluate(x, mapping, component, index_values)
        return bool(a == b)

    def __bool__(self):
        """Convert to a bool."""
        return expr_equals(self.ufl_operands[0], self.ufl_operands[1])

    __nonzero__ = __bool__


# Not associating with __ne__, the concept of equality with == is
# reserved for object equivalence for use in set and dict.
@ufl_type()
class NE(BinaryCondition):
    """Not equal condition."""

    __slots__ = ()

    def __init__(self, left, right):
        """Initialise."""
        BinaryCondition.__init__(self, "!=", left, right)

    def evaluate(self, x, mapping, component, index_values):
        """Evaluate."""
        a = self.ufl_operands[0].evaluate(x, mapping, component, index_values)
        b = self.ufl_operands[1].evaluate(x, mapping, component, index_values)
        return bool(a != b)

    def __bool__(self):
        """Convert to a bool."""
        return not expr_equals(self.ufl_operands[0], self.ufl_operands[1])

    __nonzero__ = __bool__


@ufl_type(binop="__le__")
class LE(BinaryCondition):
    """Less than or equal condition."""

    __slots__ = ()

    def __init__(self, left, right):
        """Initialise."""
        BinaryCondition.__init__(self, "<=", left, right)

    def evaluate(self, x, mapping, component, index_values):
        """Evaluate."""
        a = self.ufl_operands[0].evaluate(x, mapping, component, index_values)
        b = self.ufl_operands[1].evaluate(x, mapping, component, index_values)
        return bool(a <= b)


@ufl_type(binop="__ge__")
class GE(BinaryCondition):
    """Greater than or equal to condition."""

    __slots__ = ()

    def __init__(self, left, right):
        """Initialise."""
        BinaryCondition.__init__(self, ">=", left, right)

    def evaluate(self, x, mapping, component, index_values):
        """Evaluate."""
        a = self.ufl_operands[0].evaluate(x, mapping, component, index_values)
        b = self.ufl_operands[1].evaluate(x, mapping, component, index_values)
        return bool(a >= b)


@ufl_type(binop="__lt__")
class LT(BinaryCondition):
    """Less than condition."""

    __slots__ = ()

    def __init__(self, left, right):
        """Initialise."""
        BinaryCondition.__init__(self, "<", left, right)

    def evaluate(self, x, mapping, component, index_values):
        """Evaluate."""
        a = self.ufl_operands[0].evaluate(x, mapping, component, index_values)
        b = self.ufl_operands[1].evaluate(x, mapping, component, index_values)
        return bool(a < b)


@ufl_type(binop="__gt__")
class GT(BinaryCondition):
    """Greater than condition."""

    __slots__ = ()

    def __init__(self, left, right):
        """Initialise."""
        BinaryCondition.__init__(self, ">", left, right)

    def evaluate(self, x, mapping, component, index_values):
        """Evaluate."""
        a = self.ufl_operands[0].evaluate(x, mapping, component, index_values)
        b = self.ufl_operands[1].evaluate(x, mapping, component, index_values)
        return bool(a > b)


@ufl_type()
class AndCondition(BinaryCondition):
    """And condition."""

    __slots__ = ()

    def __init__(self, left, right):
        """Initialise."""
        BinaryCondition.__init__(self, "&&", left, right)

    def evaluate(self, x, mapping, component, index_values):
        """Evaluate."""
        a = self.ufl_operands[0].evaluate(x, mapping, component, index_values)
        b = self.ufl_operands[1].evaluate(x, mapping, component, index_values)
        return bool(a and b)


@ufl_type()
class OrCondition(BinaryCondition):
    """Or condition."""

    __slots__ = ()

    def __init__(self, left, right):
        """Initialise."""
        BinaryCondition.__init__(self, "||", left, right)

    def evaluate(self, x, mapping, component, index_values):
        """Evaluate."""
        a = self.ufl_operands[0].evaluate(x, mapping, component, index_values)
        b = self.ufl_operands[1].evaluate(x, mapping, component, index_values)
        return bool(a or b)


@ufl_type(num_ops=1)
class NotCondition(Condition):
    """Not condition."""

    __slots__ = ()

    def __init__(self, condition):
        """Initialise."""
        Condition.__init__(self, (condition,))
        if not isinstance(condition, Condition):
            raise ValueError("Expecting a condition.")

    def evaluate(self, x, mapping, component, index_values):
        """Evaluate."""
        a = self.ufl_operands[0].evaluate(x, mapping, component, index_values)
        return bool(not a)

    def __str__(self):
        """Format as a string."""
        return "!(%s)" % (str(self.ufl_operands[0]),)


@ufl_type(num_ops=3, inherit_shape_from_operand=1, inherit_indices_from_operand=1)
class Conditional(Operator):
    """Conditional expression.

    In C++ these take the format `(condition ? true_value : false_value)`.
    """

    __slots__ = ()

    def __init__(self, condition, true_value, false_value):
        """Initialise."""
        if not isinstance(condition, Condition):
            raise ValueError("Expecting condition as first argument.")
        true_value = as_ufl(true_value)
        false_value = as_ufl(false_value)
        tsh = true_value.ufl_shape
        fsh = false_value.ufl_shape
        if tsh != fsh:
            raise ValueError("Shape mismatch between conditional branches.")
        tfi = true_value.ufl_free_indices
        ffi = false_value.ufl_free_indices
        if tfi != ffi:
            raise ValueError("Free index mismatch between conditional branches.")
        if isinstance(condition, (EQ, NE)):
            if not all(
                (
                    condition.ufl_operands[0].ufl_shape == (),
                    condition.ufl_operands[0].ufl_free_indices == (),
                    condition.ufl_operands[1].ufl_shape == (),
                    condition.ufl_operands[1].ufl_free_indices == (),
                )
            ):
                raise ValueError("Non-scalar == or != is not allowed.")

        Operator.__init__(self, (condition, true_value, false_value))

    def evaluate(self, x, mapping, component, index_values):
        """Evaluate."""
        c = self.ufl_operands[0].evaluate(x, mapping, component, index_values)
        if c:
            a = self.ufl_operands[1]
        else:
            a = self.ufl_operands[2]
        return a.evaluate(x, mapping, component, index_values)

    def __str__(self):
        """Format as a string."""
        return "%s ? %s : %s" % tuple(parstr(o, self) for o in self.ufl_operands)


# --- Specific functions higher level than a conditional ---


@ufl_type(is_scalar=True, num_ops=1)
class MinValue(Operator):
    """Take the minimum of two values."""

    __slots__ = ()

    def __init__(self, left, right):
        """Initialise."""
        Operator.__init__(self, (left, right))
        if not (is_true_ufl_scalar(left) and is_true_ufl_scalar(right)):
            raise ValueError("Expecting scalar arguments.")

    def evaluate(self, x, mapping, component, index_values):
        """Evaluate."""
        a, b = self.ufl_operands
        a = a.evaluate(x, mapping, component, index_values)
        b = b.evaluate(x, mapping, component, index_values)
        try:
            res = min(a, b)
        except ValueError:
            warnings.warn("Value error in evaluation of min() of %s and %s." % self.ufl_operands)
            raise
        return res

    def __str__(self):
        """Format as a string."""
        return "min_value(%s, %s)" % self.ufl_operands


@ufl_type(is_scalar=True, num_ops=1)
class MaxValue(Operator):
    """Take the maximum of two values."""

    __slots__ = ()

    def __init__(self, left, right):
        """Initialise."""
        Operator.__init__(self, (left, right))
        if not (is_true_ufl_scalar(left) and is_true_ufl_scalar(right)):
            raise ValueError("Expecting scalar arguments.")

    def evaluate(self, x, mapping, component, index_values):
        """Evaluate."""
        a, b = self.ufl_operands
        a = a.evaluate(x, mapping, component, index_values)
        b = b.evaluate(x, mapping, component, index_values)
        try:
            res = max(a, b)
        except ValueError:
            warnings.warn("Value error in evaluation of max() of %s and %s." % self.ufl_operands)
            raise
        return res

    def __str__(self):
        """Format as a string."""
        return "max_value(%s, %s)" % self.ufl_operands



# Copyright (C) 2007 Marie Rognes
#
# This file is part of UFL.
#
# UFL is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFL. If not, see <http://www.gnu.org/licenses/>.
#
# The bilinear form a(v, u) and linear form L(v) for the Hodge Laplace
# problem using 0- and 1-forms. Intended to demonstrate use of Nedelec
# elements.
from ufl import (
    Coefficient,
    FunctionSpace,
    Mesh,
    TestFunctions,
    TrialFunctions,
    curl,
    dx,
    grad,
    inner,
    tetrahedron,
)
from ufl.finiteelement import FiniteElement, MixedElement
from ufl.pullback import covariant_piola, identity_pullback
from ufl.sobolevspace import H1, HCurl


def HodgeLaplaceGradCurl(space, fspace):
    tau, v = TestFunctions(space)
    sigma, u = TrialFunctions(space)
    f = Coefficient(fspace)

    a = (
        inner(tau, sigma) - inner(grad(tau), u) + inner(v, grad(sigma)) + inner(curl(v), curl(u))
    ) * dx
    L = inner(v, f) * dx

    return a, L


cell = tetrahedron
order = 1

GRAD = FiniteElement("Lagrange", cell, order, (), identity_pullback, H1)
CURL = FiniteElement("N1curl", cell, order, (3,), covariant_piola, HCurl)

VectorLagrange = FiniteElement("Lagrange", cell, order + 1, (3,), identity_pullback, H1)

domain = Mesh(FiniteElement("Lagrange", cell, 1, (3,), identity_pullback, H1))
space = FunctionSpace(domain, MixedElement([GRAD, CURL]))
fspace = FunctionSpace(domain, VectorLagrange)

a, L = HodgeLaplaceGradCurl(space, fspace)



__authors__ = "Martin Sandve Alnæs"
__date__ = "2008-03-12 -- 2009-01-28"

from ufl import (
    Argument,
    Coefficient,
    Coefficients,
    FacetNormal,
    FunctionSpace,
    Mesh,
    SpatialCoordinate,
    cos,
    div,
    dot,
    grad,
    i,
    inner,
    nabla_div,
    nabla_grad,
    sin,
    tan,
    triangle,
)
from ufl.algorithms import estimate_total_polynomial_degree
from ufl.finiteelement import FiniteElement, MixedElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1


def test_total_degree_estimation():
    V1 = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    V2 = FiniteElement("Lagrange", triangle, 2, (), identity_pullback, H1)
    VV = FiniteElement("Lagrange", triangle, 3, (2,), identity_pullback, H1)
    VM = MixedElement([V1, V2])

    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))

    v1_space = FunctionSpace(domain, V1)
    v2_space = FunctionSpace(domain, V2)
    vv_space = FunctionSpace(domain, VV)
    vm_space = FunctionSpace(domain, VM)

    v1 = Argument(v1_space, 2)
    v2 = Argument(v2_space, 3)
    f1, f2 = Coefficients(vm_space)
    vv = Argument(vv_space, 4)
    vu = Argument(vv_space, 5)

    x, y = SpatialCoordinate(domain)
    assert estimate_total_polynomial_degree(x) == 1
    assert estimate_total_polynomial_degree(x * y) == 2
    assert estimate_total_polynomial_degree(x**3) == 3
    assert estimate_total_polynomial_degree(x**3) == 3
    assert estimate_total_polynomial_degree((x - 1) ** 4) == 4

    assert estimate_total_polynomial_degree(vv[0]) == 3
    assert estimate_total_polynomial_degree(v2 * vv[0]) == 5
    assert estimate_total_polynomial_degree(vu[0] * vv[0]) == 6
    assert estimate_total_polynomial_degree(vu[i] * vv[i]) == 6

    assert estimate_total_polynomial_degree(v1) == 1
    assert estimate_total_polynomial_degree(v2) == 2

    # TODO: This should be 1, but 2 is expected behaviour now
    # because f1 is part of a mixed element with max degree 2.
    assert estimate_total_polynomial_degree(f1) == 2

    assert estimate_total_polynomial_degree(f2) == 2
    assert estimate_total_polynomial_degree(v2 * v1) == 3

    # TODO: This should be 2, but 3 is expected behaviour now
    # because f1 is part of a mixed element with max degree 2.
    assert estimate_total_polynomial_degree(f1 * v1) == 3

    assert estimate_total_polynomial_degree(f2 * v1) == 3
    assert estimate_total_polynomial_degree(f2 * v2 * v1) == 5

    assert estimate_total_polynomial_degree(f2 + 3) == 2
    assert estimate_total_polynomial_degree(f2 * 3) == 2
    assert estimate_total_polynomial_degree(f2**3) == 6
    assert estimate_total_polynomial_degree(f2 / 3) == 2
    assert estimate_total_polynomial_degree(f2 / v2) == 4
    assert estimate_total_polynomial_degree(f2 / (x - 1)) == 3

    assert estimate_total_polynomial_degree(v1.dx(0)) == 0
    assert estimate_total_polynomial_degree(f2.dx(0)) == 1

    assert estimate_total_polynomial_degree(f2 * v2.dx(0) * v1.dx(0)) == 2 + 1

    assert estimate_total_polynomial_degree(f2) == 2
    assert estimate_total_polynomial_degree(f2**2) == 4
    assert estimate_total_polynomial_degree(f2**3) == 6
    assert estimate_total_polynomial_degree(f2**3 * v1) == 7
    assert estimate_total_polynomial_degree(f2**3 * v1 + f1 * v1) == 7

    # Math functions of constant values are constant values
    nx, ny = FacetNormal(domain)
    e = nx**2
    for f in [sin, cos, tan, abs, lambda z: z**7]:
        assert estimate_total_polynomial_degree(f(e)) == 0

    # Based on the arbitrary chosen math function heuristics...
    heuristic_add = 2
    e = x**3
    for f in [sin, cos, tan]:
        assert estimate_total_polynomial_degree(f(e)) == 3 + heuristic_add


def test_some_compound_types():
    # NB! Although some compound types are supported here,
    # some derivatives and compounds must be preprocessed
    # prior to degree estimation. In generic code, this algorithm
    # should only be applied after preprocessing.

    etpd = estimate_total_polynomial_degree

    P2 = FiniteElement("Lagrange", triangle, 2, (), identity_pullback, H1)
    V2 = FiniteElement("Lagrange", triangle, 2, (2,), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))

    u = Coefficient(FunctionSpace(domain, P2))
    v = Coefficient(FunctionSpace(domain, V2))

    assert etpd(u.dx(0)) == 2 - 1
    assert etpd(grad(u)) == 2 - 1
    assert etpd(nabla_grad(u)) == 2 - 1
    assert etpd(div(u)) == 2 - 1

    assert etpd(v.dx(0)) == 2 - 1
    assert etpd(grad(v)) == 2 - 1
    assert etpd(nabla_grad(v)) == 2 - 1
    assert etpd(div(v)) == 2 - 1
    assert etpd(nabla_div(v)) == 2 - 1

    assert etpd(dot(v, v)) == 2 + 2
    assert etpd(inner(v, v)) == 2 + 2

    assert etpd(dot(grad(u), grad(u))) == 2 - 1 + 2 - 1
    assert etpd(inner(grad(u), grad(u))) == 2 - 1 + 2 - 1

    assert etpd(dot(grad(v), grad(v))) == 2 - 1 + 2 - 1
    assert etpd(inner(grad(v), grad(v))) == 2 - 1 + 2 - 1



"""Basic algebra operations."""
# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Anders Logg, 2008

from ufl.checks import is_true_ufl_scalar, is_ufl_scalar
from ufl.constantvalue import ComplexValue, IntValue, ScalarValue, Zero, as_ufl, zero
from ufl.core.expr import ufl_err_str
from ufl.core.operator import Operator
from ufl.core.ufl_type import ufl_type
from ufl.index_combination_utils import merge_unique_indices
from ufl.precedence import parstr
from ufl.sorting import sorted_expr

# --- Algebraic operators ---


@ufl_type(
    num_ops=2,
    inherit_shape_from_operand=0,
    inherit_indices_from_operand=0,
    binop="__add__",
    rbinop="__radd__",
)
class Sum(Operator):
    """Sum."""

    __slots__ = ()

    def __new__(cls, a, b):
        """Create a new Sum."""
        # Make sure everything is an Expr
        a = as_ufl(a)
        b = as_ufl(b)

        # Assert consistent tensor properties
        sh = a.ufl_shape
        fi = a.ufl_free_indices
        fid = a.ufl_index_dimensions
        if b.ufl_shape != sh:
            raise ValueError("Can't add expressions with different shapes.")
        if b.ufl_free_indices != fi:
            raise ValueError("Can't add expressions with different free indices.")
        if b.ufl_index_dimensions != fid:
            raise ValueError("Can't add expressions with different index dimensions.")

        # Skip adding zero
        if isinstance(a, Zero):
            return b
        elif isinstance(b, Zero):
            return a

        # Handle scalars specially and sort operands
        sa = isinstance(a, ScalarValue)
        sb = isinstance(b, ScalarValue)
        if sa and sb:
            # Apply constant propagation
            return as_ufl(a._value + b._value)
        elif sa:
            # Place scalar first
            # operands = (a, b)
            pass  # a, b = a, b
        elif sb:
            # Place scalar first
            # operands = (b, a)
            a, b = b, a
        # elif a == b:
        #    # Replace a+b with 2*foo
        #    return 2*a
        else:
            # Otherwise sort operands in a canonical order
            # operands = (b, a)
            a, b = sorted_expr((a, b))

        # construct and initialize a new Sum object
        self = Operator.__new__(cls)
        self._init(a, b)
        return self

    def _init(self, a, b):
        """Initialise."""
        self.ufl_operands = (a, b)

    def __init__(self, a, b):
        """Initialise."""
        Operator.__init__(self)

    def evaluate(self, x, mapping, component, index_values):
        """Evaluate."""
        return sum(o.evaluate(x, mapping, component, index_values) for o in self.ufl_operands)

    def __str__(self):
        """Format as a string."""
        return " + ".join([parstr(o, self) for o in self.ufl_operands])


@ufl_type(num_ops=2, binop="__mul__", rbinop="__rmul__")
class Product(Operator):
    """The product of two or more UFL objects."""

    __slots__ = ("ufl_free_indices", "ufl_index_dimensions")

    def __new__(cls, a, b):
        """Create a new product."""
        # Conversion
        a = as_ufl(a)
        b = as_ufl(b)

        # Type checking
        # Make sure everything is scalar
        if a.ufl_shape or b.ufl_shape:
            raise ValueError(
                "Product can only represent products of scalars, "
                f"got\n    {ufl_err_str(a)}\nand\n    {ufl_err_str(b)}"
            )

        # Simplification
        if isinstance(a, Zero) or isinstance(b, Zero):
            # Got any zeros? Return zero.
            fi, fid = merge_unique_indices(
                a.ufl_free_indices,
                a.ufl_index_dimensions,
                b.ufl_free_indices,
                b.ufl_index_dimensions,
            )
            return Zero((), fi, fid)
        sa = isinstance(a, ScalarValue)
        sb = isinstance(b, ScalarValue)
        if sa and sb:  # const * const = const
            # FIXME: Handle free indices like with zero? I think
            # IntValue may be index annotated now?
            return as_ufl(a._value * b._value)
        elif sa:  # 1 * b = b
            if a._value == 1:
                return b
            # a, b = a, b
        elif sb:  # a * 1 = a
            if b._value == 1:
                return a
            a, b = b, a
        # elif a == b: # a * a = a**2 # TODO: Why? Maybe just remove this?
        #    if not a.ufl_free_indices:
        #        return a**2
        else:  # a * b = b * a
            # Sort operands in a semi-canonical order
            # (NB! This is fragile! Small changes here can have large effects.)
            a, b = sorted_expr((a, b))

        # Construction
        self = Operator.__new__(cls)
        self._init(a, b)
        return self

    def _init(self, a, b):
        """Constructor, called by __new__ with already checked arguments."""
        self.ufl_operands = (a, b)

        # Extract indices
        fi, fid = merge_unique_indices(
            a.ufl_free_indices, a.ufl_index_dimensions, b.ufl_free_indices, b.ufl_index_dimensions
        )
        self.ufl_free_indices = fi
        self.ufl_index_dimensions = fid

    def __init__(self, a, b):
        """Initialise."""
        Operator.__init__(self)

    ufl_shape = ()

    def evaluate(self, x, mapping, component, index_values):
        """Evaluate."""
        ops = self.ufl_operands
        sh = self.ufl_shape
        if sh:
            if sh != ops[-1].ufl_shape:
                raise ValueError(
                    "Expecting nonscalar product operand to be the last by convention."
                )
            tmp = ops[-1].evaluate(x, mapping, component, index_values)
            ops = ops[:-1]
        else:
            tmp = 1
        for o in ops:
            tmp *= o.evaluate(x, mapping, (), index_values)
        return tmp

    def __str__(self):
        """Format as a string."""
        a, b = self.ufl_operands
        return " * ".join((parstr(a, self), parstr(b, self)))


@ufl_type(num_ops=2, inherit_indices_from_operand=0, binop="__div__", rbinop="__rdiv__")
class Division(Operator):
    """Division."""

    __slots__ = ()

    def __new__(cls, a, b):
        """Create a new Division."""
        # Conversion
        a = as_ufl(a)
        b = as_ufl(b)

        # Type checking
        # TODO: Enabled workaround for nonscalar division in __div__,
        # so maybe we can keep this assertion. Some algorithms may
        # need updating.
        if not is_ufl_scalar(a):
            raise ValueError("Expecting scalar nominator in Division.")
        if not is_true_ufl_scalar(b):
            raise ValueError("Division by non-scalar is undefined.")
        if isinstance(b, Zero):
            raise ValueError("Division by zero!")

        # Simplification
        # Simplification a/b -> a
        if isinstance(a, Zero) or (isinstance(b, ScalarValue) and b._value == 1):
            return a
        # Simplification "literal a / literal b" -> "literal value of
        # a/b". Avoiding integer division by casting to float
        if isinstance(a, ScalarValue) and isinstance(b, ScalarValue):
            try:
                return as_ufl(float(a._value) / float(b._value))
            except TypeError:
                return as_ufl(complex(a._value) / complex(b._value))
        # Simplification "a / a" -> "1"
        # if not a.ufl_free_indices and not a.ufl_shape and a == b:
        #    return as_ufl(1)

        # Construction
        self = Operator.__new__(cls)
        self._init(a, b)
        return self

    def _init(self, a, b):
        """Initialise."""
        self.ufl_operands = (a, b)

    def __init__(self, a, b):
        """Initialise."""
        Operator.__init__(self)

    ufl_shape = ()  # self.ufl_operands[0].ufl_shape

    def evaluate(self, x, mapping, component, index_values):
        """Evaluate."""
        a, b = self.ufl_operands
        a = a.evaluate(x, mapping, component, index_values)
        b = b.evaluate(x, mapping, component, index_values)
        # Avoiding integer division by casting to float
        try:
            e = float(a) / float(b)
        except TypeError:
            e = complex(a) / complex(b)
        return e

    def __str__(self):
        """Format as a string."""
        return f"{parstr(self.ufl_operands[0], self)} / {parstr(self.ufl_operands[1], self)}"


@ufl_type(num_ops=2, inherit_indices_from_operand=0, binop="__pow__", rbinop="__rpow__")
class Power(Operator):
    """Power."""

    __slots__ = ()

    def __new__(cls, a, b):
        """Create new Power."""
        # Conversion
        a = as_ufl(a)
        b = as_ufl(b)

        # Type checking
        if not is_true_ufl_scalar(a):
            raise ValueError(f"Cannot take the power of a non-scalar expression {ufl_err_str(a)}.")
        if not is_true_ufl_scalar(b):
            raise ValueError(f"Cannot raise an expression to a non-scalar power {ufl_err_str(b)}.")

        # Simplification
        if isinstance(a, ScalarValue) and isinstance(b, ScalarValue):
            return as_ufl(a._value**b._value)
        if isinstance(b, Zero):
            return IntValue(1)
        if isinstance(a, Zero) and isinstance(b, ScalarValue):
            if isinstance(b, ComplexValue):
                raise ValueError("Cannot raise zero to a complex power.")
            bf = float(b)
            if bf < 0:
                raise ValueError("Division by zero, cannot raise 0 to a negative power.")
            else:
                return zero()
        if isinstance(b, ScalarValue) and b._value == 1:
            return a

        # Construction
        self = Operator.__new__(cls)
        self._init(a, b)
        return self

    def _init(self, a, b):
        """Initialise."""
        self.ufl_operands = (a, b)

    def __init__(self, a, b):
        """Initialise."""
        Operator.__init__(self)

    ufl_shape = ()

    def evaluate(self, x, mapping, component, index_values):
        """Evalute."""
        a, b = self.ufl_operands
        a = a.evaluate(x, mapping, component, index_values)
        b = b.evaluate(x, mapping, component, index_values)
        return a**b

    def __str__(self):
        """Format as a string."""
        a, b = self.ufl_operands
        return f"{parstr(a, self)} ** {parstr(b, self)}"


@ufl_type(num_ops=1, inherit_shape_from_operand=0, inherit_indices_from_operand=0, unop="__abs__")
class Abs(Operator):
    """Absolute value."""

    __slots__ = ()

    def __new__(cls, a):
        """Create a new Abs."""
        a = as_ufl(a)

        # Simplification
        if isinstance(a, (Zero, Abs)):
            return a
        if isinstance(a, Conj):
            return Abs(a.ufl_operands[0])
        if isinstance(a, ScalarValue):
            return as_ufl(abs(a._value))

        return Operator.__new__(cls)

    def __init__(self, a):
        """Initialise."""
        Operator.__init__(self, (a,))

    def evaluate(self, x, mapping, component, index_values):
        """Evaluate."""
        a = self.ufl_operands[0].evaluate(x, mapping, component, index_values)
        return abs(a)

    def __str__(self):
        """Format as a string."""
        (a,) = self.ufl_operands
        return f"|{parstr(a, self)}|"


@ufl_type(num_ops=1, inherit_shape_from_operand=0, inherit_indices_from_operand=0)
class Conj(Operator):
    """Complex conjugate."""

    __slots__ = ()

    def __new__(cls, a):
        """Creatr a new Conj."""
        a = as_ufl(a)

        # Simplification
        if isinstance(a, (Abs, Real, Imag, Zero)):
            return a
        if isinstance(a, Conj):
            return a.ufl_operands[0]
        if isinstance(a, ScalarValue):
            return as_ufl(a._value.conjugate())

        return Operator.__new__(cls)

    def __init__(self, a):
        """Initialise."""
        Operator.__init__(self, (a,))

    def evaluate(self, x, mapping, component, index_values):
        """Evaluate."""
        a = self.ufl_operands[0].evaluate(x, mapping, component, index_values)
        return a.conjugate()

    def __str__(self):
        """Format as a string."""
        (a,) = self.ufl_operands
        return f"conj({parstr(a, self)})"


@ufl_type(num_ops=1, inherit_shape_from_operand=0, inherit_indices_from_operand=0)
class Real(Operator):
    """Real part."""

    __slots__ = ()

    def __new__(cls, a):
        """Create a new Real."""
        a = as_ufl(a)

        # Simplification
        if isinstance(a, Conj):
            a = a.ufl_operands[0]
        if isinstance(a, Zero):
            return a
        if isinstance(a, ScalarValue):
            return as_ufl(a.real())
        if isinstance(a, Real):
            a = a.ufl_operands[0]

        return Operator.__new__(cls)

    def __init__(self, a):
        """Initialise."""
        Operator.__init__(self, (a,))

    def evaluate(self, x, mapping, component, index_values):
        """Evaluate."""
        a = self.ufl_operands[0].evaluate(x, mapping, component, index_values)
        return a.real

    def __str__(self):
        """Format as a string."""
        (a,) = self.ufl_operands
        return f"Re[{parstr(a, self)}]"


@ufl_type(num_ops=1, inherit_shape_from_operand=0, inherit_indices_from_operand=0)
class Imag(Operator):
    """Imaginary part."""

    __slots__ = ()

    def __new__(cls, a):
        """Create a new Imag."""
        a = as_ufl(a)

        # Simplification
        if isinstance(a, Zero):
            return a
        if isinstance(a, (Real, Imag, Abs)):
            return Zero(a.ufl_shape, a.ufl_free_indices, a.ufl_index_dimensions)
        if isinstance(a, ScalarValue):
            return as_ufl(a.imag())

        return Operator.__new__(cls)

    def __init__(self, a):
        """Initialise."""
        Operator.__init__(self, (a,))

    def evaluate(self, x, mapping, component, index_values):
        """Evaluate."""
        a = self.ufl_operands[0].evaluate(x, mapping, component, index_values)
        return a.imag

    def __str__(self):
        """Format as a string."""
        (a,) = self.ufl_operands
        return f"Im[{parstr(a, self)}]"



__authors__ = "David Ham"
__date__ = "2014-03-04"

from math import inf

from ufl import H1, H2, L2, HCurl, HDiv, HInf, triangle
from ufl.finiteelement import FiniteElement
from ufl.pullback import contravariant_piola, covariant_piola, identity_pullback
from ufl.sobolevspace import (
    DirectionalSobolevSpace,
    SobolevSpace,  # noqa: F401
)

# Construct directional Sobolev spaces, with varying smoothness in
# spatial coordinates
H0dx0dy = DirectionalSobolevSpace((0, 0))
H1dx1dy = DirectionalSobolevSpace((1, 1))
H2dx2dy = DirectionalSobolevSpace((2, 2))
Hinfdxinfdy = DirectionalSobolevSpace((inf, inf))
H1dx = DirectionalSobolevSpace((1, 0))
H1dy = DirectionalSobolevSpace((0, 1))
H000 = DirectionalSobolevSpace((0, 0, 0))
H1dz = DirectionalSobolevSpace((0, 0, 1))
H1dh = DirectionalSobolevSpace((1, 1, 0))
H2dhH1dz = DirectionalSobolevSpace((2, 2, 1))

# TODO: Add construction of all elements with periodic table notation here.


def test_inclusion():
    assert H2 < H1  # Inclusion
    assert not H2 > H1  # Not included
    assert HDiv <= HDiv  # Reflexivity
    assert H2 < L2  # Transitivity
    assert H1 > H2
    assert L2 > H1


def test_directional_space_relations():
    assert H0dx0dy == L2
    assert H1dx1dy == H1
    assert H2dx2dy == H2
    assert H1dx1dy <= HDiv
    assert H1dx1dy <= HCurl
    assert H2dx2dy <= H1dx1dy
    assert H2dhH1dz < H1
    assert Hinfdxinfdy <= HInf
    assert Hinfdxinfdy < H2dx2dy
    assert H1dz > H2dhH1dz
    assert H1dh < L2
    assert H1dz < L2
    assert L2 > H1dx
    assert L2 > H1dy
    assert not H1dh <= HDiv
    assert not H1dh <= HCurl


def test_repr():
    assert eval(repr(H2)) == H2


def xtest_contains_mixed():
    pass  # FIXME: How to handle this?


def test_contains_l2():
    l2_elements = [
        FiniteElement("Discontinuous Lagrange", triangle, 0, (), identity_pullback, L2),
        FiniteElement("Discontinuous Lagrange", triangle, 1, (), identity_pullback, L2),
        FiniteElement("Discontinuous Lagrange", triangle, 2, (), identity_pullback, L2),
    ]
    for l2_element in l2_elements:
        assert l2_element in L2
        assert l2_element in H0dx0dy
        assert l2_element not in H1
        assert l2_element not in H1dx1dy
        assert l2_element not in HCurl
        assert l2_element not in HDiv
        assert l2_element not in H2
        assert l2_element not in H2dx2dy


def test_contains_h1():
    h1_elements = [
        # Standard Lagrange elements:
        FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1),
        FiniteElement("Lagrange", triangle, 2, (), identity_pullback, H1),
        # Some special elements:
        FiniteElement("MTW", triangle, 3, (2,), contravariant_piola, H1),
        FiniteElement("Hermite", triangle, 3, (), "custom", H1),
    ]
    for h1_element in h1_elements:
        assert h1_element in H1
        assert h1_element in H1dx1dy
        assert h1_element in HDiv
        assert h1_element in HCurl
        assert h1_element in L2
        assert h1_element in H0dx0dy
        assert h1_element not in H2
        assert h1_element not in H2dx2dy


def test_contains_h2():
    h2_elements = [
        FiniteElement("ARG", triangle, 5, (), "custom", H2),
        FiniteElement("MOR", triangle, 2, (), "custom", H2),
    ]
    for h2_element in h2_elements:
        assert h2_element in H2
        assert h2_element in H2dx2dy
        assert h2_element in H1
        assert h2_element in H1dx1dy
        assert h2_element in HDiv
        assert h2_element in HCurl
        assert h2_element in L2
        assert h2_element in H0dx0dy


def test_contains_hinf():
    hinf_elements = [FiniteElement("Real", triangle, 0, (), identity_pullback, HInf)]
    for hinf_element in hinf_elements:
        assert hinf_element in HInf
        assert hinf_element in H2
        assert hinf_element in H2dx2dy
        assert hinf_element in H1
        assert hinf_element in H1dx1dy
        assert hinf_element in HDiv
        assert hinf_element in HCurl
        assert hinf_element in L2
        assert hinf_element in H0dx0dy


def test_contains_hdiv():
    hdiv_elements = [
        FiniteElement("Raviart-Thomas", triangle, 1, (2,), contravariant_piola, HDiv),
        FiniteElement("BDM", triangle, 1, (2,), contravariant_piola, HDiv),
        FiniteElement("BDFM", triangle, 2, (2,), contravariant_piola, HDiv),
    ]
    for hdiv_element in hdiv_elements:
        assert hdiv_element in HDiv
        assert hdiv_element in L2
        assert hdiv_element in H0dx0dy
        assert hdiv_element not in H1
        assert hdiv_element not in H1dx1dy
        assert hdiv_element not in HCurl
        assert hdiv_element not in H2
        assert hdiv_element not in H2dx2dy


def test_contains_hcurl():
    hcurl_elements = [
        FiniteElement("N1curl", triangle, 1, (2,), covariant_piola, HCurl),
        FiniteElement("N2curl", triangle, 1, (2,), covariant_piola, HCurl),
    ]
    for hcurl_element in hcurl_elements:
        assert hcurl_element in HCurl
        assert hcurl_element in L2
        assert hcurl_element in H0dx0dy
        assert hcurl_element not in H1
        assert hcurl_element not in H1dx1dy
        assert hcurl_element not in HDiv
        assert hcurl_element not in H2
        assert hcurl_element not in H2dx2dy



"""Tests of domain language and attaching domains to forms."""

import pytest
from mockobjects import MockMesh

import ufl  # noqa: F401
from ufl import (
    Cell,
    Coefficient,
    Constant,
    FunctionSpace,
    Mesh,
    TestFunction,
    TrialFunction,
    dS,
    ds,
    dx,
    hexahedron,
    interval,
    quadrilateral,
    tetrahedron,
    triangle,
)
from ufl.algorithms import compute_form_data
from ufl.domain import extract_domains
from ufl.finiteelement import FiniteElement
from ufl.pullback import (
    IdentityPullback,  # noqa: F401
    identity_pullback,
)
from ufl.sobolevspace import H1

all_cells = (interval, triangle, tetrahedron, quadrilateral, hexahedron)


def test_construct_domains_from_cells():
    for cell in all_cells:
        d = cell.topological_dimension()
        Mesh(FiniteElement("Lagrange", cell, 1, (d,), identity_pullback, H1))


def test_construct_domains_with_names():
    for cell in all_cells:
        d = cell.topological_dimension()
        e = FiniteElement("Lagrange", cell, 1, (d,), identity_pullback, H1)
        D2 = Mesh(e, ufl_id=2)
        D3 = Mesh(e, ufl_id=3)
        D3b = Mesh(e, ufl_id=3)
        assert D2 != D3
        assert D3 == D3b


def test_domains_sort_by_name():
    # This ordering is rather arbitrary, but at least this shows sorting is
    # working
    domains1 = [
        Mesh(
            FiniteElement(
                "Lagrange", cell, 1, (cell.topological_dimension(),), identity_pullback, H1
            ),
            ufl_id=hash(cell.cellname()),
        )
        for cell in all_cells
    ]
    domains2 = [
        Mesh(
            FiniteElement(
                "Lagrange", cell, 1, (cell.topological_dimension(),), identity_pullback, H1
            ),
            ufl_id=hash(cell.cellname()),
        )
        for cell in sorted(all_cells)
    ]
    sdomains = sorted(domains1, key=lambda D: (D.topological_dimension(), D.ufl_cell(), D.ufl_id()))
    assert sdomains != domains1
    assert sdomains == domains2


def test_topdomain_creation():
    D = Mesh(FiniteElement("Lagrange", interval, 1, (1,), identity_pullback, H1))
    assert D.geometric_dimension() == 1
    D = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    assert D.geometric_dimension() == 2
    D = Mesh(FiniteElement("Lagrange", tetrahedron, 1, (3,), identity_pullback, H1))
    assert D.geometric_dimension() == 3


def test_cell_legacy_case():
    # Passing cell like old code does
    D = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))

    V = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    f = Coefficient(FunctionSpace(D, V))
    assert f.ufl_domains() == (D,)

    M = f * dx
    assert M.ufl_domains() == (D,)


def test_simple_domain_case():
    # Creating domain from just cell with label like new dolfin will do
    D = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1), ufl_id=3)

    V = FunctionSpace(D, FiniteElement("Lagrange", D.ufl_cell(), 1, (), identity_pullback, "H1"))
    f = Coefficient(V)
    assert f.ufl_domains() == (D,)

    M = f * dx
    assert M.ufl_domains() == (D,)


def test_creating_domains_with_coordinate_fields():  # FIXME: Rewrite for new approach
    # Definition of higher order domain, element, coefficient, form

    # Mesh with P2 representation of coordinates
    cell = triangle
    P2 = FiniteElement("Lagrange", cell, 2, (2,), identity_pullback, H1)
    domain = Mesh(P2)

    # Piecewise linear function space over quadratic mesh
    element = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
    V = FunctionSpace(domain, element)

    f = Coefficient(V)
    M = f * dx
    assert f.ufl_domains() == (domain,)
    assert M.ufl_domains() == (domain,)

    # Test the gymnastics that dolfin will have to go through
    domain2 = Mesh(P2, ufl_id=domain.ufl_id())
    V2 = FunctionSpace(domain2, eval(repr(V.ufl_element())))
    f2 = Coefficient(V2, count=f.count())
    assert f == f2
    assert domain == domain2
    assert V == V2


def test_join_domains():
    from ufl.domain import join_domains

    mesh7 = MockMesh(7)
    mesh8 = MockMesh(8)
    triangle = Cell("triangle")
    xa = FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)
    xb = FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)

    # Equal domains are joined
    assert 1 == len(
        join_domains(
            [
                Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1), ufl_id=7),
                Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1), ufl_id=7),
            ]
        )
    )
    assert 1 == len(
        join_domains(
            [
                Mesh(
                    FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1),
                    ufl_id=7,
                    cargo=mesh7,
                ),
                Mesh(
                    FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1),
                    ufl_id=7,
                    cargo=mesh7,
                ),
            ]
        )
    )
    assert 1 == len(join_domains([Mesh(xa, ufl_id=3), Mesh(xa, ufl_id=3)]))

    # Different domains are not joined
    assert 2 == len(
        join_domains(
            [
                Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)),
                Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)),
            ]
        )
    )
    assert 2 == len(
        join_domains(
            [
                Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1), ufl_id=7),
                Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1), ufl_id=8),
            ]
        )
    )
    assert 2 == len(
        join_domains(
            [
                Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1), ufl_id=7),
                Mesh(
                    FiniteElement("Lagrange", quadrilateral, 1, (2,), identity_pullback, H1),
                    ufl_id=8,
                ),
            ]
        )
    )
    assert 2 == len(join_domains([Mesh(xa, ufl_id=7), Mesh(xa, ufl_id=8)]))
    assert 2 == len(join_domains([Mesh(xa), Mesh(xb)]))

    # Incompatible coordinates require labeling
    xc = Coefficient(
        FunctionSpace(
            Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)),
            FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1),
        )
    )
    xd = Coefficient(
        FunctionSpace(
            Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)),
            FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1),
        )
    )
    with pytest.raises(BaseException):
        join_domains([Mesh(xc), Mesh(xd)])

    # Incompatible data is checked if and only if the domains are the same
    assert 2 == len(
        join_domains(
            [
                Mesh(
                    FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1),
                    ufl_id=7,
                    cargo=mesh7,
                ),
                Mesh(
                    FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1),
                    ufl_id=8,
                    cargo=mesh8,
                ),
            ]
        )
    )
    assert 2 == len(
        join_domains(
            [
                Mesh(
                    FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1),
                    ufl_id=7,
                    cargo=mesh7,
                ),
                Mesh(
                    FiniteElement("Lagrange", quadrilateral, 1, (2,), identity_pullback, H1),
                    ufl_id=8,
                    cargo=mesh8,
                ),
            ]
        )
    )
    # Geometric dimensions must match
    with pytest.raises(BaseException):
        join_domains(
            [
                Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)),
                Mesh(FiniteElement("Lagrange", triangle, 1, (3,), identity_pullback, H1)),
            ]
        )
    with pytest.raises(BaseException):
        join_domains(
            [
                Mesh(
                    FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1),
                    ufl_id=7,
                    cargo=mesh7,
                ),
                Mesh(
                    FiniteElement("Lagrange", triangle, 1, (3,), identity_pullback, H1),
                    ufl_id=8,
                    cargo=mesh8,
                ),
            ]
        )
    # Cargo and mesh ids must match
    with pytest.raises(BaseException):
        Mesh(
            FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1),
            ufl_id=7,
            cargo=mesh8,
        )

    # Nones are removed
    assert 2 == len(
        join_domains(
            [
                None,
                Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1), ufl_id=3),
                None,
                Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1), ufl_id=3),
                None,
                Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1), ufl_id=4),
            ]
        )
    )
    assert 2 == len(
        join_domains(
            [
                Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1), ufl_id=7),
                None,
                Mesh(
                    FiniteElement("Lagrange", quadrilateral, 1, (2,), identity_pullback, H1),
                    ufl_id=8,
                ),
            ]
        )
    )
    assert None not in join_domains(
        [
            Mesh(FiniteElement("Lagrange", triangle, 1, (3,), identity_pullback, H1), ufl_id=7),
            None,
            Mesh(FiniteElement("Lagrange", tetrahedron, 1, (3,), identity_pullback, H1), ufl_id=8),
        ]
    )


def test_everywhere_integrals_with_backwards_compatibility():
    D = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))

    V = FunctionSpace(D, FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1))
    f = Coefficient(V)

    a = f * dx
    (ida,) = compute_form_data(a).integral_data

    # Check some integral data
    assert ida.integral_type == "cell"
    assert len(ida.subdomain_id) == 1
    assert ida.subdomain_id[0] == "otherwise"
    assert ida.metadata == {}

    # Integrands are not equal because of renumbering
    itg1 = ida.integrals[0].integrand()
    itg2 = a.integrals()[0].integrand()
    assert type(itg1) is type(itg2)
    assert itg1.ufl_element() == itg2.ufl_element()


def test_merge_sort_integral_data():
    D = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))

    V = FunctionSpace(D, FiniteElement("CG", triangle, 1, (), identity_pullback, H1))

    u = Coefficient(V)
    c = Constant(D)
    a = c * dS((2, 4)) + u * dx + u * ds + 2 * u * dx(3) + 2 * c * dS + 2 * u * dx((1, 4))
    form_data = compute_form_data(a, do_append_everywhere_integrals=False).integral_data
    assert len(form_data) == 5

    # Check some integral data
    assert form_data[0].integral_type == "cell"
    assert len(form_data[0].subdomain_id) == 1
    assert form_data[0].subdomain_id[0] == "otherwise"
    assert form_data[0].metadata == {}

    assert form_data[1].integral_type == "cell"
    assert len(form_data[1].subdomain_id) == 3
    assert form_data[1].subdomain_id[0] == 1
    assert form_data[1].subdomain_id[1] == 3
    assert form_data[1].subdomain_id[2] == 4
    assert form_data[1].metadata == {}

    assert form_data[2].integral_type == "exterior_facet"
    assert len(form_data[2].subdomain_id) == 1
    assert form_data[2].subdomain_id[0] == "otherwise"
    assert form_data[2].metadata == {}

    assert form_data[3].integral_type == "interior_facet"
    assert len(form_data[3].subdomain_id) == 1
    assert form_data[3].subdomain_id[0] == "otherwise"
    assert form_data[3].metadata == {}

    assert form_data[4].integral_type == "interior_facet"
    assert len(form_data[4].subdomain_id) == 2
    assert form_data[4].subdomain_id[0] == 2
    assert form_data[4].subdomain_id[1] == 4
    assert form_data[4].metadata == {}


def test_extract_domains():
    "Test that the domains are extracted properly from a mixed-domain expression"

    # Create domains of different topological dimensions
    gdim = 2
    cell_type_0 = triangle
    cell_type_1 = interval
    dom_0 = Mesh(FiniteElement("Lagrange", cell_type_0, 1, (gdim,), identity_pullback, H1))
    dom_1 = Mesh(FiniteElement("Lagrange", cell_type_1, 1, (gdim,), identity_pullback, H1))

    # Define some finite element spaces
    k = 1
    ele_type = "Lagrange"
    ele_0 = FiniteElement(ele_type, cell_type_0, k, (), identity_pullback, H1)
    ele_1 = FiniteElement(ele_type, cell_type_1, k, (), identity_pullback, H1)

    V_0 = FunctionSpace(dom_0, ele_0)
    V_1 = FunctionSpace(dom_1, ele_1)

    # Create test and trial functions
    u = TrialFunction(V_0)
    v = TestFunction(V_1)

    # Create a mixed-domain expression
    expr = u * v

    domains = extract_domains(expr)

    assert domains[0] == dom_1
    assert domains[1] == dom_0



import gc
import sys

from ufl import (
    Coefficient,
    Constant,
    FunctionSpace,
    Mesh,
    TestFunction,
    TrialFunction,
    dx,
    grad,
    inner,
    triangle,
)
from ufl.algorithms import replace_terminal_data, strip_terminal_data
from ufl.core.ufl_id import attach_ufl_id
from ufl.core.ufl_type import UFLObject
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1

MIN_REF_COUNT = 2
"""The minimum value returned by sys.getrefcount."""


@attach_ufl_id
class AugmentedMesh(Mesh, UFLObject):
    def __init__(self, *args, data):
        super().__init__(*args)
        self.data = data


class AugmentedFunctionSpace(FunctionSpace, UFLObject):
    def __init__(self, *args, data):
        super().__init__(*args)
        self.data = data


class AugmentedCoefficient(Coefficient):
    def __init__(self, *args, data):
        super().__init__(*args)
        self.data = data


class AugmentedConstant(Constant):
    def __init__(self, *args, data):
        super().__init__(*args)
        self.data = data


def test_strip_form_arguments_strips_data_refs():
    mesh_data = object()
    fs_data = object()
    coeff_data = object()
    const_data = object()

    # Sanity check
    assert sys.getrefcount(mesh_data) == MIN_REF_COUNT
    assert sys.getrefcount(fs_data) == MIN_REF_COUNT
    assert sys.getrefcount(coeff_data) == MIN_REF_COUNT
    assert sys.getrefcount(const_data) == MIN_REF_COUNT

    cell = triangle
    domain = AugmentedMesh(
        FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1), data=mesh_data
    )
    element = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
    V = AugmentedFunctionSpace(domain, element, data=fs_data)

    v = TestFunction(V)
    u = TrialFunction(V)
    f = AugmentedCoefficient(V, data=coeff_data)
    k = AugmentedConstant(V, data=const_data)

    form = k * f * inner(grad(v), grad(u)) * dx

    # Remove extraneous references
    del cell, domain, element, V, v, u, f, k

    assert sys.getrefcount(mesh_data) == MIN_REF_COUNT + 1
    assert sys.getrefcount(fs_data) == MIN_REF_COUNT + 1
    assert sys.getrefcount(coeff_data) == MIN_REF_COUNT + 1
    assert sys.getrefcount(const_data) == MIN_REF_COUNT + 1

    stripped_form, mapping = strip_terminal_data(form)

    del form, mapping
    gc.collect()  # This is needed to update the refcounts

    assert sys.getrefcount(mesh_data) == MIN_REF_COUNT
    assert sys.getrefcount(fs_data) == MIN_REF_COUNT
    assert sys.getrefcount(coeff_data) == MIN_REF_COUNT
    assert sys.getrefcount(const_data) == MIN_REF_COUNT


def test_strip_form_arguments_does_not_change_form():
    mesh_data = object()
    fs_data = object()
    coeff_data = object()
    const_data = object()

    cell = triangle
    domain = AugmentedMesh(
        FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1), data=mesh_data
    )
    element = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
    V = AugmentedFunctionSpace(domain, element, data=fs_data)

    v = TestFunction(V)
    u = TrialFunction(V)
    f = AugmentedCoefficient(V, data=coeff_data)
    k = AugmentedConstant(V, data=const_data)

    form = k * f * inner(grad(v), grad(u)) * dx
    stripped_form, mapping = strip_terminal_data(form)

    assert stripped_form.signature() == form.signature()
    assert replace_terminal_data(stripped_form, mapping) == form



"""This module collects algorithms and utility functions operating on UFL objects."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Anders Logg, 2008-2009.


# FIXME: Clean up this to become a more official set of supported
#        algorithms.  Currently contains too much stuff that's not
#        recommended to use. The __all__ list below is a start based
#        on grepping of other FEniCS code for ufl.algorithm imports.

__all__ = [
    "estimate_total_polynomial_degree",
    "sort_elements",
    "compute_form_data",
    "preprocess_form",
    "apply_transformer",
    "ReuseTransformer",
    "load_ufl_file",
    "Transformer",
    "MultiFunction",
    "extract_unique_elements",
    "extract_type",
    "extract_elements",
    "extract_sub_elements",
    "expand_indices",
    "replace",
    "expand_derivatives",
    "extract_coefficients",
    "extract_base_form_operators",
    "strip_variables",
    "strip_terminal_data",
    "replace_terminal_data",
    "post_traversal",
    "change_to_reference_grad",
    "validate_form",
    "FormSplitter",
    "extract_arguments",
    "compute_form_adjoint",
    "compute_form_action",
    "compute_energy_norm",
    "compute_form_lhs",
    "compute_form_rhs",
    "compute_form_functional",
    "compute_form_signature",
    "compute_form_arities",
    "tree_format",
    "read_ufl_file",
    "load_forms",
]

from ufl.algorithms.ad import expand_derivatives
from ufl.algorithms.analysis import (
    extract_arguments,
    extract_base_form_operators,
    extract_coefficients,
    extract_elements,
    extract_sub_elements,
    extract_type,
    extract_unique_elements,
    sort_elements,
)
from ufl.algorithms.change_to_reference import change_to_reference_grad
from ufl.algorithms.checks import validate_form
from ufl.algorithms.compute_form_data import compute_form_data, preprocess_form
from ufl.algorithms.estimate_degrees import estimate_total_polynomial_degree
from ufl.algorithms.expand_indices import expand_indices
from ufl.algorithms.formfiles import load_forms, load_ufl_file, read_ufl_file
from ufl.algorithms.formsplitter import FormSplitter
from ufl.algorithms.formtransformations import (
    compute_energy_norm,
    compute_form_action,
    compute_form_adjoint,
    compute_form_arities,
    compute_form_functional,
    compute_form_lhs,
    compute_form_rhs,
)
from ufl.algorithms.replace import replace
from ufl.algorithms.signature import compute_form_signature
from ufl.algorithms.strip_terminal_data import replace_terminal_data, strip_terminal_data
from ufl.algorithms.transformer import (
    ReuseTransformer,
    Transformer,
    apply_transformer,
    strip_variables,
)
from ufl.corealg.multifunction import MultiFunction
from ufl.corealg.traversal import post_traversal
from ufl.utils.formatting import tree_format
